# Copyright (c) 2011-2019, Manfred Moitzi
# License: MIT License
import pytest
from ezdxf.lldxf.tagger import internal_tag_compiler
from ezdxf.document import Drawing
from ezdxf import DXFValueError, decode_base64


def test_dxfversion_1():
    doc = Drawing.from_tags(internal_tag_compiler(TEST_HEADER))
    assert "AC1009" == doc.dxfversion


@pytest.fixture(scope="module")
def dwg_r12():
    return Drawing.new("AC1009")


def test_dxfversion_2(dwg_r12):
    assert "AC1009" == dwg_r12.dxfversion


def test_acad_release(dwg_r12):
    assert "R12" == dwg_r12.acad_release


def test_get_layer(dwg_r12):
    layer = dwg_r12.layers.get("0")
    assert "0" == layer.dxf.name


def test_error_getting_not_existing_layer(dwg_r12):
    with pytest.raises(DXFValueError):
        layer = dwg_r12.layers.get("TEST_NOT_EXISTING_LAYER")


def test_create_layer(dwg_r12):
    layer = dwg_r12.layers.new("TEST_NEW_LAYER")
    assert "TEST_NEW_LAYER" == layer.dxf.name


def test_error_adding_existing_layer(dwg_r12):
    with pytest.raises(DXFValueError):
        layer = dwg_r12.layers.new("0")


def test_has_layer(dwg_r12):
    assert "0" in dwg_r12.layers


def test_has_not_layer(dwg_r12):
    assert "TEST_LAYER_NOT_EXISTS" not in dwg_r12.layers


def test_removing_layer(dwg_r12):
    dwg_r12.layers.new("TEST_NEW_LAYER_2")
    assert "TEST_NEW_LAYER_2" in dwg_r12.layers
    dwg_r12.layers.remove("TEST_NEW_LAYER_2")
    assert "TEST_NEW_LAYER_2" not in dwg_r12.layers


def test_error_removing_not_existing_layer(dwg_r12):
    with pytest.raises(DXFValueError):
        dwg_r12.layers.remove("TEST_LAYER_NOT_EXISTS")


@pytest.fixture(scope="module")
def dwg_r2000():
    return Drawing.new("AC1015")


def test_r2000_dxfversion(dwg_r2000):
    assert "AC1015" == dwg_r2000.dxfversion


def test_r2000_acad_release(dwg_r2000):
    assert "R2000" == dwg_r2000.acad_release


@pytest.fixture
def min_r12():
    return Drawing.from_tags(internal_tag_compiler(MINIMALISTIC_DXF12))


def test_min_r12_header_section(min_r12):
    assert hasattr(min_r12, "header")
    assert min_r12.header["$ACADVER"] == "AC1009"
    assert min_r12.header["$DWGCODEPAGE"] == "ANSI_1252"


def test_min_r12_layers_table(min_r12):
    assert hasattr(min_r12, "layers")
    assert len(min_r12.layers) == 2
    assert "0" in min_r12.layers
    assert "Defpoints" in min_r12.layers


def test_min_r12_styles_table(min_r12):
    assert hasattr(min_r12, "styles")
    assert len(min_r12.styles) == 1
    assert "Standard" in min_r12.styles


def test_min_r12_linetypes_table(min_r12):
    assert hasattr(min_r12, "linetypes")
    assert len(min_r12.linetypes) == 3
    assert "continuous" in min_r12.linetypes
    assert "ByLayer" in min_r12.linetypes
    assert "ByBlock" in min_r12.linetypes


def test_min_r12_blocks_section(min_r12):
    assert hasattr(min_r12, "blocks")
    assert len(min_r12.blocks) == 2
    assert "*Model_Space" in min_r12.blocks
    assert "*Paper_Space" in min_r12.blocks


def test_min_r12_entity_section(min_r12):
    assert hasattr(min_r12, "entities")
    assert len(min_r12.entities) == 0


def test_chain_layout_and_block(dwg_r12, dwg_r2000):
    for dwg in (dwg_r12, dwg_r2000):
        msp = dwg.modelspace()
        line_msp = msp.add_line((0, 0), (1, 1))
        blk = dwg.blocks.new("TEST_CHAIN")
        line_blk = blk.add_line((0, 0), (1, 1))

        handles = list(e.dxf.handle for e in dwg.chain_layouts_and_blocks())
        # check for unique handles
        assert len(handles) == len(set(handles))

        check = {line_msp.dxf.handle, line_blk.dxf.handle}
        assert check.intersection(handles) == check


def test_base64_encoding_r12(dwg_r12):
    data = dwg_r12.encode_base64()
    doc = decode_base64(data)
    assert doc.acad_release == "R12"


def test_base64_encoding_r2000(dwg_r2000):
    data = dwg_r2000.encode_base64()
    doc = decode_base64(data)
    assert doc.acad_release == "R2000"


def test_set_drawing_units(dwg_r12):
    dwg_r12.units = 6
    assert dwg_r12.header["$INSUNITS"] == 6
    dwg_r12.units = 5
    assert dwg_r12.header["$INSUNITS"] == 5


MINIMALISTIC_DXF12 = """  0
SECTION
  2
ENTITIES
  0
ENDSEC
  0
EOF
"""

TEST_HEADER = """  0
SECTION
  2
HEADER
  9
$ACADVER
  1
AC1009
  9
$DWGCODEPAGE
  3
ANSI_1252
  9
$HANDSEED
  5
FF
  0
ENDSEC
  0
SECTION
  2
ENTITIES
  0
ENDSEC
  0
EOF
"""

TESTCOPY = """  0
SECTION
  2
HEADER
  9
$ACADVER
  1
AC1018
  9
$DWGCODEPAGE
  3
ANSI_1252
  9
$TDUPDATE
 40
0.
  9
$HANDSEED
  5
FF
  0
ENDSEC
  0
SECTION
  2
OBJECTS
  0
ENDSEC
  0
SECTION
  2
FANTASYSECTION
  1
everything should be copied
  0
ENDSEC
  0
SECTION
  2
ALPHASECTION
  1
everything should be copied
  0
ENDSEC
  0
SECTION
  2
OMEGASECTION
  1
everything should be copied
  0
ENDSEC
  0
EOF
"""



# Copyright (C) 2011-2022, Manfred Moitzi
# License: MIT License
import pytest
from ezdxf.entities.layer import Layer
from ezdxf.lldxf.const import DXFValueError


@pytest.fixture
def layer():
    return Layer.new(handle="FFFF")


def test_get_handle(layer):
    assert "FFFF" == layer.dxf.handle


def test_get_name(layer):
    assert "0" == layer.dxf.name


def test_get_flags(layer):
    assert layer.dxf.flags == 0


def test_get_color(layer):
    assert layer.dxf.color == 7


def test_get_linetype(layer):
    assert layer.dxf.linetype.upper() == "CONTINUOUS"


def test_set_name(layer):
    layer.dxf.name = "MOZMAN"
    assert "MOZMAN" == layer.dxf.name


def test_set_color(layer):
    layer.dxf.color = "1"
    assert 1 == layer.dxf.color


def test_set_color_2(layer):
    layer.set_color(-1)
    assert 1 == layer.get_color()


def test_set_color_for_off_layer(layer):
    layer.set_color(7)
    layer.off()
    assert 7 == layer.get_color()
    assert -7 == layer.dxf.color


def test_color_as_property(layer):
    layer.color = 7
    layer.off()
    assert 7 == layer.color
    assert -7 == layer.dxf.color


def test_is_locked(layer):
    layer.lock()
    assert layer.is_locked() is True


def test_is_not_locked(layer):
    assert layer.is_locked() is False


def test_is_on(layer):
    assert layer.is_on() is True


def test_is_off(layer):
    layer.dxf.color = -100
    assert layer.is_off() is True


def test_is_frozen(layer):
    assert layer.is_frozen() is False


def test_freeze(layer):
    layer.freeze()
    assert layer.is_frozen() is True
    assert 1 == layer.dxf.flags


def test_thaw(layer):
    layer.dxf.flags = 1
    assert layer.is_frozen() is True
    layer.thaw()
    assert layer.is_frozen() is False
    assert 0 == layer.dxf.flags


def test_invald_layer_name():
    with pytest.raises(DXFValueError):
        Layer.new("FFFF", dxfattribs={"name": "Layer/"})


def test_set_true_color_as_rgb(layer):
    layer.rgb = (10, 10, 10)
    assert layer.dxf.true_color == 657930


def test_get_true_color_as_rgb(layer):
    layer.dxf.true_color = 657930
    assert layer.rgb == (10, 10, 10)


def test_get_default_description(layer):
    assert layer.description == ""


def test_get_default_description_at_existing_xdata(layer):
    layer.set_xdata("mozman", [(1000, "test")])
    assert layer.description == ""


def test_description_for_unusual_xdata_structure(layer):
    # Just one group code 1000 tag - not reproducible with BricsCAD
    layer.set_xdata("AcAecLayerStandard", [(1000, "test")])
    assert layer.description == ""

    # or empty XDATA section - not reproducible with BricsCAD
    layer.set_xdata("AcAecLayerStandard", [])
    assert layer.description == ""


def test_set_description(layer):
    layer.description = "my Layer"
    assert layer.description == "my Layer"

    # test DXF internals
    assert "AcAecLayerStandard" in layer.xdata


def test_replace_description(layer):
    layer.description = "my Layer"
    assert layer.description == "my Layer"
    layer.description = "new Description"
    assert layer.description == "new Description"


def test_get_default_transparency(layer):
    assert layer.transparency == 0


def test_fully_transparent_layer(layer):
    layer.set_xdata("AcCmTransparency", [(1071, 0x02000000)])
    assert layer.transparency == 1.0


def test_half_transparent_layer(layer):
    layer.set_xdata("AcCmTransparency", [(1071, 0x0200007F)])
    assert round(layer.transparency, 2) == 0.5


def test_opaque_layer(layer):
    layer.set_xdata("AcCmTransparency", [(1071, 0x020000FF)])
    assert layer.transparency == 0.0


def test_invalid_transparency_returns_opaque(layer):
    # The flag 0x02000000 has to be set for a valid transparency
    layer.set_xdata("AcCmTransparency", [(1071, 0)])
    assert layer.transparency == 0.0


def test_transparency_byblock_returns_opaque(layer):
    # Transparency BYBLOCK (0x01000000) make no sense for a layer!?
    layer.set_xdata("AcCmTransparency", [(1071, 0x01000000)])
    assert layer.transparency == 0.0


def test_set_transparency(layer):
    layer.transparency = 0.11
    assert round(layer.transparency, 2) == 0.11

    # test DXF internals
    assert "AcCmTransparency" in layer.xdata


def test_replace_transparency(layer):
    layer.transparency = 0.11
    assert round(layer.transparency, 2) == 0.11
    layer.transparency = 0.77
    assert round(layer.transparency, 2) == 0.77


MALFORMED_LAYER = """0
LAYER
2
LY_EZDXF
5
FEFE
100
AcDbLayerTableRecord
70
0
62
7
6
CONTINUOUS
100
AcDbSymbolTableRecord
"""


def test_malformed_layer():
    layer = Layer.from_text(MALFORMED_LAYER)
    assert layer.dxf.name == "LY_EZDXF"
    assert layer.dxf.handle == "FEFE"
    assert layer.dxf.color == 7
    assert layer.dxf.linetype == "CONTINUOUS"



# Copyright (c) 2019-2024 Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import TYPE_CHECKING, Optional
from typing_extensions import Self
from ezdxf.lldxf import validator
from ezdxf.lldxf.attributes import (
    DXFAttr,
    DXFAttributes,
    DefSubclass,
    XType,
    RETURN_DEFAULT,
    group_code_mapping,
    merge_group_code_mappings,
)
from ezdxf.lldxf.const import (
    SUBCLASS_MARKER,
    DXF12,
    DXF2000,
    MODEL_SPACE_R12,
    PAPER_SPACE_R12,
    MODEL_SPACE_R2000,
    PAPER_SPACE_R2000,
)
from ezdxf.math import NULLVEC
from .dxfentity import base_class, SubclassProcessor, DXFEntity
from .factory import register_entity
from ezdxf.audit import Auditor, AuditError

if TYPE_CHECKING:
    from ezdxf.entities import DXFNamespace
    from ezdxf.lldxf.tagwriter import AbstractTagWriter
    from ezdxf import xref

__all__ = ["Block", "EndBlk"]

acdb_entity = DefSubclass(
    "AcDbEntity",
    {
        # No auto fix for invalid layer names!
        "layer": DXFAttr(8, default="0", validator=validator.is_valid_layer_name),
        "paperspace": DXFAttr(
            67,
            default=0,
            optional=True,
            validator=validator.is_integer_bool,
            fixer=RETURN_DEFAULT,
        ),
    },
)
acdb_entity_group_codes = group_code_mapping(acdb_entity)

acdb_block_begin = DefSubclass(
    "AcDbBlockBegin",
    {
        "name": DXFAttr(2, validator=validator.is_valid_block_name),
        # The 2nd name with group code 3 is handled internally, and is not an
        # explicit DXF attribute.
        "description": DXFAttr(4, default="", dxfversion=DXF2000, optional=True),
        # Flags:
        # 0 = Indicates none of the following flags apply
        # 1 = This is an anonymous block generated by hatching, associative
        #     dimensioning, other internal operations, or an application
        # 2 = This block has non-constant attribute definitions (this bit is not set
        #     if the block has any attribute definitions that are constant, or has
        #     no attribute definitions at all)
        # 4 = This block is an external reference (xref)
        # 8 = This block is an xref overlay
        # 16 = This block is externally dependent
        # 32 = This is a resolved external reference, or dependent of an external
        #      reference (ignored on input)
        # 64 = This definition is a referenced external reference (ignored on input)
        "flags": DXFAttr(70, default=0),
        "base_point": DXFAttr(10, xtype=XType.any_point, default=NULLVEC),
        "xref_path": DXFAttr(1, default=""),
    },
)
acdb_block_begin_group_codes = group_code_mapping(acdb_block_begin)
merged_block_begin_group_codes = merge_group_code_mappings(
    acdb_entity_group_codes, acdb_block_begin_group_codes
)

MODEL_SPACE_R2000_LOWER = MODEL_SPACE_R2000.lower()
MODEL_SPACE_R12_LOWER = MODEL_SPACE_R12.lower()
PAPER_SPACE_R2000_LOWER = PAPER_SPACE_R2000.lower()
PAPER_SPACE_R12_LOWER = PAPER_SPACE_R12.lower()


@register_entity
class Block(DXFEntity):
    """DXF BLOCK entity"""

    DXFTYPE = "BLOCK"
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_block_begin)

    # Block entity flags:
    # This is an anonymous block generated by hatching, associative
    # dimensioning, other internal operations, or an application:
    ANONYMOUS = 1

    # This block has non-constant attribute definitions (this bit is not set
    # if the block has any attribute definitions that are constant, or has no
    # attribute definitions at all):
    NON_CONSTANT_ATTRIBUTES = 2

    # This block is an external reference:
    XREF = 4

    # This block is an xref overlay:
    XREF_OVERLAY = 8

    # This block is externally dependent:
    EXTERNAL = 16

    # This is a resolved external reference, or dependent of an external reference:
    RESOLVED = 32

    # This definition is a referenced external reference:
    REFERENCED = 64

    def load_dxf_attribs(
        self, processor: Optional[SubclassProcessor] = None
    ) -> DXFNamespace:
        """Loading interface. (internal API)"""
        dxf = super().load_dxf_attribs(processor)
        if processor is None:
            return dxf
        processor.simple_dxfattribs_loader(dxf, merged_block_begin_group_codes)
        if processor.r12:
            if dxf.name is None:
                dxf.name = ""
            name = dxf.name.lower()
            if name == MODEL_SPACE_R12_LOWER:
                dxf.name = MODEL_SPACE_R2000
            elif name == PAPER_SPACE_R12_LOWER:
                dxf.name = PAPER_SPACE_R2000
        return dxf

    def export_entity(self, tagwriter: AbstractTagWriter) -> None:
        """Export entity specific data as DXF tags."""
        super().export_entity(tagwriter)

        if tagwriter.dxfversion > DXF12:
            tagwriter.write_tag2(SUBCLASS_MARKER, acdb_entity.name)
        if self.dxf.hasattr("paperspace"):
            tagwriter.write_tag2(67, 1)
        self.dxf.export_dxf_attribs(tagwriter, "layer")
        if tagwriter.dxfversion > DXF12:
            tagwriter.write_tag2(SUBCLASS_MARKER, acdb_block_begin.name)

        name = self.dxf.name
        if tagwriter.dxfversion == DXF12:
            # export modelspace and paperspace with leading '$' instead of '*'
            if name.lower() == MODEL_SPACE_R2000_LOWER:
                name = MODEL_SPACE_R12
            elif name.lower() == PAPER_SPACE_R2000_LOWER:
                name = PAPER_SPACE_R12

        tagwriter.write_tag2(2, name)
        self.dxf.export_dxf_attribs(tagwriter, ["flags", "base_point"])
        tagwriter.write_tag2(3, name)
        self.dxf.export_dxf_attribs(tagwriter, ["xref_path", "description"])

    @property
    def is_layout_block(self) -> bool:
        """Returns ``True`` if this is a :class:`~ezdxf.layouts.Modelspace` or
        :class:`~ezdxf.layouts.Paperspace` block definition.
        """
        name = self.dxf.name.lower()
        return name.startswith("*model_space") or name.startswith("*paper_space")

    @property
    def is_anonymous(self) -> bool:
        """Returns ``True`` if this is an anonymous block generated by
        hatching, associative dimensioning, other internal operations, or an
        application.

        """
        return self.get_flag_state(Block.ANONYMOUS)

    @property
    def is_xref(self) -> bool:
        """Returns ``True`` if bock is an external referenced file."""
        return self.get_flag_state(Block.XREF)

    @property
    def is_xref_overlay(self) -> bool:
        """Returns ``True`` if bock is an external referenced overlay file."""
        return self.get_flag_state(Block.XREF_OVERLAY)

    def audit(self, auditor: Auditor):
        owner_handle = self.dxf.get("owner")
        if owner_handle is None:  # invalid owner handle - IGNORE
            return
        owner = auditor.entitydb.get(owner_handle)
        if owner is None:  # invalid owner entity - IGNORE
            return
        owner_name = owner.dxf.get("name", "").upper()
        block_name = self.dxf.get("name", "").upper()
        if owner_name != block_name:
            auditor.add_error(
                AuditError.BLOCK_NAME_MISMATCH,
                f"{str(self)} name '{block_name}' and {str(owner)} name '{owner_name}' mismatch",
            )

    def map_resources(self, clone: Self, mapping: xref.ResourceMapper) -> None:
        """Translate resources from self to the copied entity."""
        assert isinstance(clone, Block)
        super().map_resources(clone, mapping)
        clone.dxf.name = mapping.get_block_name(self.dxf.name)


acdb_block_end = DefSubclass("AcDbBlockEnd", {})


@register_entity
class EndBlk(DXFEntity):
    """DXF ENDBLK entity"""

    DXFTYPE = "ENDBLK"
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_block_end)

    def load_dxf_attribs(
        self, processor: Optional[SubclassProcessor] = None
    ) -> DXFNamespace:
        """Loading interface. (internal API)"""
        dxf = super().load_dxf_attribs(processor)
        if processor:
            processor.simple_dxfattribs_loader(dxf, acdb_entity_group_codes)  # type: ignore
        return dxf

    def export_entity(self, tagwriter: AbstractTagWriter) -> None:
        """Export entity specific data as DXF tags."""
        super().export_entity(tagwriter)

        if tagwriter.dxfversion > DXF12:
            tagwriter.write_tag2(SUBCLASS_MARKER, acdb_entity.name)
        if self.dxf.hasattr("paperspace"):
            tagwriter.write_tag2(67, 1)
        self.dxf.export_dxf_attribs(tagwriter, "layer")
        if tagwriter.dxfversion > DXF12:
            tagwriter.write_tag2(SUBCLASS_MARKER, acdb_block_end.name)



# Copyright (c) 2021-2022, Manfred Moitzi
# License: MIT License
from __future__ import annotations
import enum
from typing import NamedTuple, Union

from ezdxf.math import Vec2, Vec3

__all__ = [
    "Command",
    "AnyCurve",
    "PathElement",
    "LineTo",
    "Curve3To",
    "Curve4To",
    "MoveTo",
]


@enum.unique
class Command(enum.IntEnum):
    LINE_TO = 1  # (LINE_TO, end vertex)
    CURVE3_TO = 2  # (CURVE3_TO, end vertex, ctrl) quadratic bezier
    CURVE4_TO = 3  # (CURVE4_TO, end vertex, ctrl1, ctrl2) cubic bezier
    MOVE_TO = 4  # (MOVE_TO, end vertex), creates a gap and starts a sub-path


class LineTo(NamedTuple):
    end: Vec2|Vec3

    @property
    def type(self):
        return Command.LINE_TO


class MoveTo(NamedTuple):
    end: Vec2|Vec3

    @property
    def type(self):
        return Command.MOVE_TO


class Curve3To(NamedTuple):
    end: Vec2|Vec3
    ctrl: Vec2|Vec3

    @property
    def type(self):
        return Command.CURVE3_TO


class Curve4To(NamedTuple):
    end: Vec2|Vec3
    ctrl1: Vec2|Vec3
    ctrl2: Vec2|Vec3

    @property
    def type(self):
        return Command.CURVE4_TO


AnyCurve = (Command.CURVE3_TO, Command.CURVE4_TO)
PathElement = Union[LineTo, Curve3To, Curve4To, MoveTo]



# Purpose: create sierpinski pyramid geometry
# Copyright (c) 2016-2022 Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import TYPE_CHECKING, Iterable, Sequence, Iterator, Optional
import math
from ezdxf.math import Vec3, UVec, Matrix44, UCS
from ezdxf.render.mesh import MeshVertexMerger, MeshTransformer

if TYPE_CHECKING:
    from ezdxf.eztypes import GenericLayoutType

HEIGHT4 = 1.0 / math.sqrt(2.0)  # pyramid4 height (* length)
HEIGHT3 = math.sqrt(6.0) / 3.0  # pyramid3 height (* length)

DY1_FACTOR = math.tan(math.pi / 6.0) / 2.0  # inner circle radius
DY2_FACTOR = 0.5 / math.cos(math.pi / 6.0)  # outer circle radius


class SierpinskyPyramid:
    """
    Args:
        location: location of base center as (x, y, z) tuple
        length: side length
        level: subdivide level
        sides: sides of base geometry

    """

    def __init__(
        self,
        location: UVec = (0.0, 0.0, 0.0),
        length: float = 1.0,
        level: int = 1,
        sides: int = 4,
    ):
        self.sides = sides
        self.pyramid_definitions = sierpinsky_pyramid(
            location=location, length=length, level=level, sides=sides
        )

    def vertices(self) -> Iterator[list[Vec3]]:
        """Yields the pyramid vertices as list of :class:`ezdxf.math.Vec3`."""
        for location, length in self.pyramid_definitions:
            yield self._calc_vertices(location, length)

    __iter__ = vertices

    def _calc_vertices(
        self, location: UVec, length: float
    ) -> list[Vec3]:
        """
        Calculates the pyramid vertices.

        Args:
            location: location of the pyramid as center point of the base
            length: pyramid side length

        Returns: list of :class:`ezdxf.math.Vec3`

        """
        len2 = length / 2.0
        x, y, z = location
        if self.sides == 4:
            return [
                Vec3(x - len2, y - len2, z),
                Vec3(x + len2, y - len2, z),
                Vec3(x + len2, y + len2, z),
                Vec3(x - len2, y + len2, z),
                Vec3(x, y, z + length * HEIGHT4),
            ]
        elif self.sides == 3:
            dy1 = length * DY1_FACTOR
            dy2 = length * DY2_FACTOR
            return [
                Vec3(x - len2, y - dy1, z),
                Vec3(x + len2, y - dy1, z),
                Vec3(x, y + dy2, z),
                Vec3(x, y, z + length * HEIGHT3),
            ]
        else:
            raise ValueError("sides has to be 3 or 4.")

    def faces(self) -> list[Sequence[int]]:
        """Returns list of pyramid faces. All pyramid vertices have the same
        order, so one faces list fits them all.

        """
        if self.sides == 4:
            return [(3, 2, 1, 0), (0, 1, 4), (1, 2, 4), (2, 3, 4), (3, 0, 4)]
        elif self.sides == 3:
            return [(2, 1, 0), (0, 1, 3), (1, 2, 3), (2, 0, 3)]
        else:
            raise ValueError("sides has to be 3 or 4.")

    def render(
        self,
        layout: GenericLayoutType,
        merge: bool = False,
        dxfattribs=None,
        matrix: Optional[Matrix44] = None,
        ucs: Optional[UCS] = None,
    ) -> None:
        """Renders the sierpinsky pyramid into layout, set `merge` to ``True``
        for rendering the whole sierpinsky pyramid into one MESH entity, set
        `merge` to ``False`` for individual pyramids as MESH entities.

        Args:
            layout: DXF target layout
            merge: ``True`` for one MESH entity, ``False`` for individual MESH
                entities per pyramid
            dxfattribs: DXF attributes for the MESH entities
            matrix: apply transformation matrix at rendering
            ucs: apply UCS at rendering

        """
        if merge:
            mesh = self.mesh()
            mesh.render_mesh(
                layout, dxfattribs=dxfattribs, matrix=matrix, ucs=ucs
            )
        else:
            for pyramid in self.pyramids():
                pyramid.render_mesh(layout, dxfattribs, matrix=matrix, ucs=ucs)

    def pyramids(self) -> Iterable[MeshTransformer]:
        """Yields all pyramids of the sierpinsky pyramid as individual
        :class:`MeshTransformer` objects.
        """
        faces = self.faces()
        for vertices in self:
            mesh = MeshTransformer()
            mesh.add_mesh(vertices=vertices, faces=faces)
            yield mesh

    def mesh(self) -> MeshTransformer:
        """Returns geometry as one :class:`MeshTransformer` object."""
        faces = self.faces()
        mesh = MeshVertexMerger()
        for vertices in self:
            mesh.add_mesh(vertices=vertices, faces=faces)
        return MeshTransformer.from_builder(mesh)


def sierpinsky_pyramid(
    location=(0.0, 0.0, 0.0),
    length: float = 1.0,
    level: int = 1,
    sides: int = 4,
) -> list[tuple[Vec3, float]]:
    """Build a Sierpinski pyramid.

    Args:
        location: base center point of the pyramid
        length: base length of the pyramid
        level: recursive building levels, has to 1 or bigger
        sides: 3 or 4 sided pyramids supported

    Returns: list of pyramid vertices

    """
    location = Vec3(location)
    level = int(level)
    if level < 1:
        raise ValueError("level has to be 1 or bigger.")
    pyramids = _sierpinsky_pyramid(location, length, sides)
    for _ in range(level - 1):
        next_level_pyramids = []
        for location, length in pyramids:
            next_level_pyramids.extend(
                _sierpinsky_pyramid(location, length, sides)
            )
        pyramids = next_level_pyramids
    return pyramids


def _sierpinsky_pyramid(
    location: Vec3, length: float = 1.0, sides: int = 4
) -> list[tuple[Vec3, float]]:
    if sides == 3:
        return sierpinsky_pyramid_3(location, length)
    elif sides == 4:
        return sierpinsky_pyramid_4(location, length)
    else:
        raise ValueError("sides has to be 3 or 4.")


def sierpinsky_pyramid_4(
    location: Vec3, length: float = 1.0
) -> list[tuple[Vec3, float]]:
    """Build a 4-sided Sierpinski pyramid. Pyramid height = length of the base
    square!

    Args:
        location: base center point of the pyramid
        length: base length of the pyramid

    Returns: list of (location, length) tuples, representing the sierpinski pyramid

    """
    len2 = length / 2
    len4 = length / 4
    x, y, z = location
    return [
        (Vec3(x - len4, y - len4, z), len2),
        (Vec3(x + len4, y - len4, z), len2),
        (Vec3(x - len4, y + len4, z), len2),
        (Vec3(x + len4, y + len4, z), len2),
        (Vec3(x, y, z + len2 * HEIGHT4), len2),
    ]


def sierpinsky_pyramid_3(
    location: Vec3, length: float = 1.0
) -> list[tuple[Vec3, float]]:
    """Build a 3-sided Sierpinski pyramid (tetraeder).

    Args:
        location: base center point of the pyramid
        length: base length of the pyramid

    Returns: list of (location, length) tuples, representing the sierpinski pyramid

    """
    dy1 = length * DY1_FACTOR * 0.5
    dy2 = length * DY2_FACTOR * 0.5
    len2 = length / 2
    len4 = length / 4
    x, y, z = location
    return [
        (Vec3(x - len4, y - dy1, z), len2),
        (Vec3(x + len4, y - dy1, z), len2),
        (Vec3(x, y + dy2, z), len2),
        (Vec3(x, y, z + len2 * HEIGHT3), len2),
    ]



# Copyright (c) 2019-2024 Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import TYPE_CHECKING, Iterator
import math
import numpy as np

from ezdxf.math import (
    Vec3,
    Matrix44,
    ConstructionArc,
    arc_angle_span_deg,
)
from ezdxf.math.transformtools import OCSTransform

from ezdxf.lldxf.attributes import (
    DXFAttr,
    DXFAttributes,
    DefSubclass,
    group_code_mapping,
    merge_group_code_mappings,
)
from ezdxf.lldxf.const import DXF12, SUBCLASS_MARKER
from .dxfentity import base_class
from .dxfgfx import acdb_entity
from .circle import acdb_circle, Circle, merged_circle_group_codes
from .factory import register_entity

if TYPE_CHECKING:
    from ezdxf.lldxf.tagwriter import AbstractTagWriter

__all__ = ["Arc"]

acdb_arc = DefSubclass(
    "AcDbArc",
    {
        "start_angle": DXFAttr(50, default=0),
        "end_angle": DXFAttr(51, default=360),
    },
)

acdb_arc_group_codes = group_code_mapping(acdb_arc)
merged_arc_group_codes = merge_group_code_mappings(
    merged_circle_group_codes, acdb_arc_group_codes
)


@register_entity
class Arc(Circle):
    """DXF ARC entity"""

    DXFTYPE = "ARC"
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_circle, acdb_arc)
    MERGED_GROUP_CODES = merged_arc_group_codes

    def export_entity(self, tagwriter: AbstractTagWriter) -> None:
        """Export entity specific data as DXF tags."""
        super().export_entity(tagwriter)
        # AcDbEntity export is done by parent class
        # AcDbCircle export is done by parent class
        if tagwriter.dxfversion > DXF12:
            tagwriter.write_tag2(SUBCLASS_MARKER, acdb_arc.name)
        self.dxf.export_dxf_attribs(tagwriter, ["start_angle", "end_angle"])

    @property
    def start_point(self) -> Vec3:
        """Returns the start point of the arc in :ref:`WCS`, takes the :ref:`OCS` into
        account.
        """
        v = list(self.vertices([self.dxf.start_angle]))
        return v[0]

    @property
    def end_point(self) -> Vec3:
        """Returns the end point of the arc in :ref:`WCS`, takes the :ref:`OCS` into
        account.
        """
        v = list(self.vertices([self.dxf.end_angle]))
        return v[0]

    def angles(self, num: int) -> Iterator[float]:
        """Yields `num` angles from start- to end angle in degrees in counter-clockwise
        orientation. All angles are normalized in the range from [0, 360).
        """
        if num < 2:
            raise ValueError("num >= 2")
        start = self.dxf.start_angle % 360
        stop = self.dxf.end_angle % 360
        if stop <= start:
            stop += 360
        for angle in np.linspace(start, stop, num=num, endpoint=True):
            yield angle % 360

    def flattening(self, sagitta: float) -> Iterator[Vec3]:
        """Approximate the arc by vertices in :ref:`WCS`, the argument `sagitta`_
        defines the maximum distance from the center of an arc segment to the center of
        its chord.

        .. _sagitta: https://en.wikipedia.org/wiki/Sagitta_(geometry)
        """
        arc = self.construction_tool()
        ocs = self.ocs()
        elevation = Vec3(self.dxf.center).z
        if ocs.transform:
            to_wcs = ocs.points_to_wcs
        else:
            to_wcs = Vec3.generate

        yield from to_wcs(Vec3(p.x, p.y, elevation) for p in arc.flattening(sagitta))

    def transform(self, m: Matrix44) -> Arc:
        """Transform ARC entity by transformation matrix `m` inplace.
        Raises ``NonUniformScalingError()`` for non-uniform scaling.
        """
        ocs = OCSTransform(self.dxf.extrusion, m)
        super()._transform(ocs)
        s: float = self.dxf.start_angle
        e: float = self.dxf.end_angle
        if not math.isclose(arc_angle_span_deg(s, e), 360.0):
            (
                self.dxf.start_angle,
                self.dxf.end_angle,
            ) = ocs.transform_ccw_arc_angles_deg(s, e)
        self.post_transform(m)
        return self

    def construction_tool(self) -> ConstructionArc:
        """Returns the 2D construction tool :class:`ezdxf.math.ConstructionArc` but the
        extrusion vector is ignored.
        """
        dxf = self.dxf
        return ConstructionArc(
            dxf.center,
            dxf.radius,
            dxf.start_angle,
            dxf.end_angle,
        )

    def apply_construction_tool(self, arc: ConstructionArc) -> Arc:
        """Set ARC data from the construction tool :class:`ezdxf.math.ConstructionArc`
        but the extrusion vector is ignored.
        """
        dxf = self.dxf
        dxf.center = Vec3(arc.center)
        dxf.radius = arc.radius
        dxf.start_angle = arc.start_angle
        dxf.end_angle = arc.end_angle
        return self  # floating interface



# Copyright (C) 2016-2020, Manfred Moitzi
# License: MIT License
import pytest
import codecs
from ezdxf.lldxf.encoding import dxf_backslash_replace, encode
from pathlib import Path

# setup DXF unicode encoder -> '\U+nnnn'
codecs.register_error("dxfreplace", dxf_backslash_replace)


def test_ascii_encoding():
    assert b"123\\U+6539" == encode("123改", "ascii")


@pytest.mark.parametrize(
    ["s", "e"],
    [
        ("300\n\udcb7\udc9e\udcff\n", b"300\n\xb7\x9e\xff\n"),
        ("123改", b"123\\U+6539"),
    ],
)
def test_surrogate_escape_support_in_dxf_replace_encoder(s, e):
    assert e == encode(s, "ascii")


@pytest.mark.parametrize("n", [0, 1, 2])
def test_XRECORD_handling_of_dxf_replace_encoder(n):
    XRECORD = Path(__file__).parent / f"XRECORD_{n}.bin"
    with open(XRECORD, "rb") as f:
        data = f.read()
    s = data.decode("utf8", errors="surrogateescape")
    result = encode(s, encoding="utf8")
    assert data == result



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License

import pytest
from ezdxf import colors

# DXF transparency value 0x020000TT
# 0x00 = fully transparent
# 0xff = opaque


@pytest.mark.parametrize("t_int, t_float", [
    (0x02000000, 1.0),
    (0x0200007F, 0.5),
    (0x020000FF, 0.0),
    (0, 1.0),
    (127, 0.5),
    (255, 0.0),
])
def test_transparency_to_float(t_int, t_float):
    assert round(colors.transparency2float(t_int), 2) == t_float


@pytest.mark.parametrize("t_int, t_float", [
    (0x02000000, 1.0),
    (0x0200007F, 0.5),
    (0x020000FF, 0.0),
])
def test_float_to_transparency(t_int, t_float):
    assert colors.float2transparency(t_float) == t_int


if __name__ == '__main__':
    pytest.main([__file__])



# Copyright (c) 2024, Manfred Moitzi
# License: MIT License
import pytest

from ezdxf.math._construct import world_mercator_to_gps, gps_to_world_mercator
from ezdxf.math import Vec2
from ezdxf.acc import USE_C_EXT

if USE_C_EXT:
    try:
        from ezdxf.acc.construct import world_mercator_to_gps as cy_world_mercator_to_gps
        from ezdxf.acc.construct import gps_to_world_mercator as cy_gps_to_world_mercator
    except ImportError:
        USE_C_EXT = False

DATA = [
    [Vec2(15, 47), Vec2(1669792.36, 5910809.62)],
    [Vec2(-15, 47), Vec2(-1669792.36, 5910809.62)],
    [Vec2(15, -47), Vec2(1669792.36, -5910809.62)],
    [Vec2(-15, -47), Vec2(-1669792.36, -5910809.62)],
    [Vec2(0, 0), Vec2(0, 0)],
]


@pytest.mark.parametrize("deg, coords", DATA)
def test_common_WGS84_projection_cpython(deg: Vec2, coords: Vec2):
    projected = Vec2(gps_to_world_mercator(deg.x, deg.y))
    assert projected.round(2).isclose(coords)
    # inverse projection
    assert Vec2(world_mercator_to_gps(projected.x, projected.y)).isclose(deg)


@pytest.mark.skipif(USE_C_EXT is False, reason="Cython implementation not available")
@pytest.mark.parametrize("deg, coords", DATA)
def test_common_WGS84_projection_cython(deg: Vec2, coords: Vec2):
    projected = Vec2(cy_gps_to_world_mercator(deg.x, deg.y))
    assert projected.round(2).isclose(coords)
    # inverse projection
    assert Vec2(cy_world_mercator_to_gps(projected.x, projected.y)).isclose(deg)


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2010-2020 Manfred Moitzi
# License: MIT License
import pytest
import random

from ezdxf.math import (
    cubic_bezier_interpolation,
    Vec3,
    Vec2,
    Bezier3P,
    quadratic_to_cubic_bezier,
    Bezier4P,
    have_bezier_curves_g1_continuity,
    bezier_to_bspline,
    split_bezier,
    quadratic_bezier_from_3p,
    close_vectors,
    cubic_bezier_bbox,
    quadratic_bezier_bbox,
    intersection_ray_cubic_bezier_2d,
)


def test_vertex_interpolation():
    points = [(0, 0), (3, 1), (5, 3), (0, 8)]
    result = list(cubic_bezier_interpolation(points))
    assert len(result) == 3
    c1, c2, c3 = result
    p = c1.control_points
    assert p[0].isclose((0, 0))
    assert p[1].isclose((0.9333333333333331, 0.3111111111111111))
    assert p[2].isclose((1.8666666666666663, 0.6222222222222222))
    assert p[3].isclose((3, 1))

    p = c2.control_points
    assert p[0].isclose((3, 1))
    assert p[1].isclose((4.133333333333334, 1.3777777777777778))
    assert p[2].isclose((5.466666666666667, 1.822222222222222))
    assert p[3].isclose((5, 3))

    p = c3.control_points
    assert p[0].isclose((5, 3))
    assert p[1].isclose((4.533333333333333, 4.177777777777778))
    assert p[2].isclose((2.2666666666666666, 6.088888888888889))
    assert p[3].isclose((0, 8))


def test_invalid_bezier_interpolation():
    """At least 3 points are required."""
    assert len(list(cubic_bezier_interpolation([(0, 0)]))) == 0
    assert len(list(cubic_bezier_interpolation([(0, 0), (1, 0)]))) == 0


def test_quadratic_to_cubic_bezier():
    r = random.Random(0)

    def random_vec() -> Vec3:
        return Vec3(r.uniform(-10, 10), r.uniform(-10, 10), r.uniform(-10, 10))

    for i in range(1000):
        quadratic = Bezier3P((random_vec(), random_vec(), random_vec()))
        quadratic_approx = list(quadratic.approximate(10))
        cubic = quadratic_to_cubic_bezier(quadratic)
        cubic_approx = list(cubic.approximate(10))

        assert len(quadratic_approx) == len(cubic_approx)
        for p1, p2 in zip(quadratic_approx, cubic_approx):
            assert p1.isclose(p2)


# G1 continuity: normalized end-tangent == normalized start-tangent of next curve
B1 = Bezier4P(Vec2.list([(0, 0), (1, 1), (2, 1), (3, 0)]))

# B1/B2 has G1 continuity:
B2 = Bezier4P(Vec2.list([(3, 0), (4, -1), (5, -1), (6, 0)]))

# B1/B3 has no G1 continuity:
B3 = Bezier4P(Vec2.list([(3, 0), (4, 1), (5, 1), (6, 0)]))

# B1/B4 G1 continuity off tolerance:
B4 = Bezier4P(Vec2.list([(3, 0), (4, -1.03), (5, -1.0), (6, 0)]))

# B1/B5 has a gap between B1 end and B5 start:
B5 = Bezier4P(Vec2.list([(4, 0), (5, -1), (6, -1), (7, 0)]))


def test_g1_continuity_for_bezier_curves():
    assert have_bezier_curves_g1_continuity(B1, B2) is True
    assert have_bezier_curves_g1_continuity(B1, B3) is False
    assert (
        have_bezier_curves_g1_continuity(B1, B4, g1_tol=1e-4) is False
    ), "should be outside of tolerance "
    assert (
        have_bezier_curves_g1_continuity(B1, B5) is False
    ), "end- and start point should match"


D1 = Bezier4P(Vec2.list([(0, 0), (1, 1), (3, 0), (3, 0)]))
D2 = Bezier4P(Vec2.list([(3, 0), (3, 0), (5, -1), (6, 0)]))


def test_g1_continuity_for_degenerated_bezier_curves():
    assert have_bezier_curves_g1_continuity(D1, B2) is False
    assert have_bezier_curves_g1_continuity(B1, D2) is False
    assert have_bezier_curves_g1_continuity(D1, D2) is False


@pytest.mark.parametrize("curve", [D1, D2])
def test_flatten_degenerated_bezier_curves(curve):
    # Degenerated Bezier curves behave like regular curves!
    assert len(list(curve.flattening(0.1))) > 4


@pytest.mark.parametrize(
    "b1,b2",
    [
        (B1, B2),  # G1 continuity, the common case
        (B1, B3),  # without G1 continuity is also a regular B-spline
        (B1, B5),  # regular B-spline, but first control point of B5 is lost
    ],
    ids=["G1", "without G1", "gap"],
)
def test_bezier_curves_to_bspline(b1, b2):
    bspline = bezier_to_bspline([b1, b2])
    # Remove duplicate control point between two adjacent curves:
    expected = list(b1.control_points) + list(b2.control_points)[1:]
    assert bspline.degree == 3, "should be a cubic B-spline"
    assert bspline.control_points == tuple(expected)


def test_quality_of_bezier_to_bspline_conversion_1():
    # This test shows the close relationship between cubic Bézier- and
    # cubic B-spline curves.
    points0 = B1.approximate(10)
    points1 = bezier_to_bspline([B1]).approximate(10)
    for p0, p1 in zip(points0, points1):
        assert p0.isclose(p1) is True, "conversion should be perfect"


def test_quality_of_bezier_to_bspline_conversion_2():
    # This test shows the close relationship between cubic Bézier- and
    # cubic B-spline curves.
    # Remove duplicate point between the two curves:
    points0 = list(B1.approximate(10)) + list(B2.approximate(10))[1:]
    points1 = bezier_to_bspline([B1, B2]).approximate(20)
    for p0, p1 in zip(points0, points1):
        assert p0.isclose(p1) is True, "conversion should be perfect"


def test_bezier_curves_to_bspline_error():
    with pytest.raises(ValueError):
        bezier_to_bspline([])  # one or more curves expected


class TestSplitBezier:
    @pytest.fixture
    def points3(self):
        return Vec2.list([(0, 0), (0, 1), (1.5, 0.75), (2, 2)])

    @pytest.mark.parametrize("t", [-1, 2])
    def test_t_validation(self, points3, t):
        with pytest.raises(ValueError):
            split_bezier(points3, t)

    def test_control_point_validation(self):
        with pytest.raises(ValueError):
            split_bezier([Vec2(0, 0)], 0.5)

    def test_split_cubic_bezier(self, points3):
        left, right = split_bezier(points3, 0.5)
        assert (
            close_vectors(
                left,
                [(0.0, 0.0), (0.0, 0.5), (0.375, 0.6875), (0.8125, 0.90625)],
            )
            is True
        )

        assert (
            close_vectors(
                right,
                [(2.0, 2.0), (1.75, 1.375), (1.25, 1.125), (0.8125, 0.90625)],
            )
            is True
        )


def test_quadratic_bezier_from_3_points():
    qbez = quadratic_bezier_from_3p((0, 0), (3, 2), (6, 0))
    assert qbez.point(0.5).isclose((3, 2))


def test_cubic_bezier_from_3_points():
    cbez = quadratic_bezier_from_3p((0, 0), (3, 2), (6, 0))
    assert cbez.point(0.5).isclose((3, 2))


class TestBezierCurveBoundingBox:
    def test_linear_curve(self):
        bbox = cubic_bezier_bbox(Bezier4P(Vec2.list([(0, 0), (1, 1), (2, 2), (3, 3)])))
        assert bbox.extmin == (0, 0, 0)
        assert bbox.extmax == (3, 3, 0)

    def test_reverse_linear_curve(self):
        bbox = cubic_bezier_bbox(
            Bezier4P(Vec2.list([(3, 3), (2, 2), (-2, -2), (-3, -3)]))
        )
        assert bbox.extmin == (-3, -3, 0)
        assert bbox.extmax == (3, 3, 0)

    def test_cubic_bezier_curve_with_one_extrema(self):
        curve = Bezier4P(Vec2.list([(0, 0), (0, 1), (2, 1), (2, 0)]))
        bbox = cubic_bezier_bbox(curve)
        assert bbox.extmax.y == pytest.approx(0.75)

    def test_cubic_bezier_curve_with_two_extrema(self):
        curve = Bezier4P(Vec2.list([(0, 0), (0, 1), (2, -1), (2, 0)]))
        bbox = cubic_bezier_bbox(curve)
        assert bbox.extmin.y == pytest.approx(-0.28867513459481287)
        assert bbox.extmax.y == pytest.approx(+0.28867513459481287)

    def test_closed_3d_cubic_bezier_curve(self):
        curve = Bezier4P(Vec3.list([(0, 0, -1), (2, 3, 0), (-2, 3, 0), (0, 0, -1)]))
        bbox = cubic_bezier_bbox(curve)
        assert bbox.extmin.x == pytest.approx(-0.5773502691896258)
        assert bbox.extmin.z == pytest.approx(-1.0)
        assert bbox.extmax.x == pytest.approx(+0.5773502691896258)
        assert bbox.extmax.y == pytest.approx(+2.25)
        assert bbox.extmax.z == pytest.approx(-0.25)

    def test_quadratic_bezier_curve_box(self):
        curve = Bezier3P(Vec2.list([(0, 0), (1, 1), (2, 0)]))
        bbox = quadratic_bezier_bbox(curve)
        assert bbox.extmax.y == pytest.approx(0.5)


class TestRayCubicBezierCurve2dIntersection:
    @pytest.fixture(scope="class")
    def curve(self):
        return Bezier4P(Vec2.list([(0, -2), (2, 6), (4, -6), (6, 2)]))

    def test_no_intersection(self, curve):
        assert len(intersection_ray_cubic_bezier_2d((0, -6), (1, -6), curve)) == 0

    def test_one_intersection_point(self, curve):
        points = intersection_ray_cubic_bezier_2d((3, -6), (3, 6), curve)
        assert len(points) == 1
        assert points[0].isclose((3, 0))

    def test_two_intersection_points(self, curve):
        points = intersection_ray_cubic_bezier_2d((-1.4, -2.5), (7.1, 3.9), curve)
        assert len(points) == 2
        expected = (
            (0.18851028511733303, -1.3039451970881237),
            (2.5249135145844264, 0.4552289992165126),
        )
        assert all(p.isclose(e) for e, p in zip(expected, points)) is True

    def test_three_intersection_points(self, curve):
        points = intersection_ray_cubic_bezier_2d((0, 0), (1, 0), curve)
        assert len(points) == 3
        expected = (
            (0.6762099922755492, 0.0),
            (3.0, 0.0),
            (5.323790007724451, 0.0),
        )
        assert all(p.isclose(e) for e, p in zip(expected, points)) is True

    def test_collinear_ray_and_curve(self):
        curve = Bezier4P(Vec2.list([(0, 0), (1, 0), (2, 0), (3, 0)]))
        ip = intersection_ray_cubic_bezier_2d((0, 0), (1, 0), curve)
        assert len(ip) == 1
        assert ip[0].isclose((0, 0))  # ???

    @pytest.mark.parametrize("x", [0, 0.5, 1, 3])
    def test_linear_ray_and_curve(self, x):
        curve = Bezier4P(Vec2.list([(0, 0), (1, 0), (2, 0), (3, 0)]))
        # ray defined in +y direction
        ip = intersection_ray_cubic_bezier_2d((x, -1), (x, 0), curve)
        assert len(ip) == 1
        assert ip[0].isclose((x, 0))
        # ray defined in -y direction
        ip = intersection_ray_cubic_bezier_2d((x, 2), (x, 1), curve)
        assert len(ip) == 1
        assert ip[0].isclose((x, 0))



# Copyright (c) 2023, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import TYPE_CHECKING, TypeVar, NamedTuple, Optional
from copy import deepcopy
import logging
from ezdxf.lldxf.const import DXFError

if TYPE_CHECKING:
    from ezdxf.entities import DXFEntity


__all__ = ["CopyStrategy", "CopySettings", "CopyNotSupported", "default_copy"]


T = TypeVar("T", bound="DXFEntity")


class CopyNotSupported(DXFError):
    pass


class CopySettings(NamedTuple):
    reset_handles: bool = True
    copy_extension_dict: bool = True
    copy_xdata: bool = True
    copy_appdata: bool = True
    copy_reactors: bool = False
    copy_proxy_graphic: bool = True
    set_source_of_copy: bool = True

    # The processing of copy errors of linked entities has to be done in the
    # copy_data() method by the entity itself!
    ignore_copy_errors_in_linked_entities: bool = True


class LogMessage(NamedTuple):
    message: str
    level: int = logging.WARNING
    entity: Optional[DXFEntity] = None


class CopyStrategy:
    log: list[LogMessage] = []

    def __init__(self, settings: CopySettings) -> None:
        self.settings = settings

    def copy(self, entity: T) -> T:
        """Entity copy for usage in the same document or as virtual entity.

        This copy is NOT stored in the entity database and does NOT reside in any
        layout, block, table or objects section!
        """
        settings = self.settings
        clone = entity.__class__()
        doc = entity.doc
        clone.doc = doc
        clone.dxf = entity.dxf.copy(clone)
        if settings.reset_handles:
            clone.dxf.reset_handles()

        if settings.copy_extension_dict:
            xdict = entity.extension_dict
            if xdict is not None and doc is not None and xdict.is_alive:
                # Error handling of unsupported entities in the extension dictionary is
                # done by the underlying DICTIONARY entity.
                clone.extension_dict = xdict.copy(self)

        if settings.copy_reactors and entity.reactors is not None:
            clone.reactors = entity.reactors.copy()

        if settings.copy_proxy_graphic:
            clone.proxy_graphic = entity.proxy_graphic  # immutable bytes

        # if appdata contains handles, they are treated as shared resources
        if settings.copy_appdata:
            clone.appdata = deepcopy(entity.appdata)

        # if xdata contains handles, they are treated as shared resources
        if settings.copy_xdata:
            clone.xdata = deepcopy(entity.xdata)

        if settings.set_source_of_copy:
            clone.set_source_of_copy(entity)

        entity.copy_data(clone, copy_strategy=self)
        return clone

    @classmethod
    def add_log_message(
        cls, msg: str, level: int = logging.WARNING, entity: Optional[DXFEntity] = None
    ) -> None:
        cls.log.append(LogMessage(msg, level, entity))

    @classmethod
    def clear_log_message(cls) -> None:
        cls.log.clear()


# same strategy as DXFEntity.copy() of v1.1.3
default_copy = CopyStrategy(CopySettings())



# Copyright (c) 2021-2022, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import Iterable, Iterator, Sequence
from .mtext import MText, MTextColumns

__all__ = [
    "make_static_columns_r2000",
    "make_dynamic_auto_height_columns_r2000",
    "make_dynamic_manual_height_columns_r2000",
    "make_static_columns_r2018",
    "make_dynamic_auto_height_columns_r2018",
    "make_dynamic_manual_height_columns_r2018",
]

COLUMN_BREAK = "\\N"


def add_column_breaks(content: Iterable[str]) -> Iterator[str]:
    content = list(content)
    for c in content[:-1]:
        if not c.endswith(COLUMN_BREAK):
            c += COLUMN_BREAK
        yield c
    yield content[-1]  # last column without a column break


def make_static_columns_r2000(
    content: Sequence[str],
    width: float,
    gutter_width: float,
    height: float,
    dxfattribs=None,
) -> MText:
    if len(content) < 1:
        raise ValueError("no content")
    columns = MTextColumns.new_static_columns(
        len(content), width, gutter_width, height
    )
    mtext = MText.new(dxfattribs=dxfattribs)
    mtext.setup_columns(columns, linked=True)
    content = list(add_column_breaks(content))
    mtext.text = content[0]
    for mt, c in zip(columns.linked_columns, content[1:]):
        mt.text = c
    return mtext


def make_dynamic_auto_height_columns_r2000(
    content: str,
    width: float,
    gutter_width: float,
    height: float,
    count: int = 1,
    dxfattribs=None,
) -> MText:
    if not content:
        raise ValueError("no content")
    mtext = MText.new(dxfattribs=dxfattribs)
    mtext.dxf.width = width
    columns = MTextColumns.new_dynamic_auto_height_columns(
        count, width, gutter_width, height
    )
    set_dynamic_columns_content(content, mtext, columns)
    return mtext


def make_dynamic_manual_height_columns_r2000(
    content: str,
    width: float,
    gutter_width: float,
    heights: Sequence[float],
    dxfattribs=None,
) -> MText:
    if not content:
        raise ValueError("no content")
    mtext = MText.new(dxfattribs=dxfattribs)
    mtext.dxf.width = width
    columns = MTextColumns.new_dynamic_manual_height_columns(
        width, gutter_width, heights
    )
    set_dynamic_columns_content(content, mtext, columns)
    return mtext


def set_dynamic_columns_content(
    content: str, mtext: MText, columns: MTextColumns
):
    mtext.setup_columns(columns, linked=True)
    # temp. hack: assign whole content to the main column
    mtext.text = content
    # for mt, c in zip(mtext.columns.linked_columns, content[1:]):
    #    mt.text = c
    return mtext


# DXF version R2018


def make_static_columns_r2018(
    content: Sequence[str],
    width: float,
    gutter_width: float,
    height: float,
    dxfattribs=None,
) -> MText:
    if len(content) < 1:
        raise ValueError("no content")
    columns = MTextColumns.new_static_columns(
        len(content), width, gutter_width, height
    )
    mtext = MText.new(dxfattribs=dxfattribs)
    mtext.setup_columns(columns, linked=False)
    mtext.text = "".join(add_column_breaks(content))
    return mtext


def make_dynamic_auto_height_columns_r2018(
    content: str,
    width: float,
    gutter_width: float,
    height: float,
    count: int,
    dxfattribs=None,
) -> MText:
    columns = MTextColumns.new_dynamic_auto_height_columns(
        count, width, gutter_width, height
    )
    return _make_dynamic_columns_r2018(content, columns, dxfattribs or {})


def make_dynamic_manual_height_columns_r2018(
    content: str,
    width: float,
    gutter_width: float,
    heights: Sequence[float],
    dxfattribs=None,
) -> MText:
    columns = MTextColumns.new_dynamic_manual_height_columns(
        width, gutter_width, heights
    )
    return _make_dynamic_columns_r2018(content, columns, dxfattribs or {})


def _make_dynamic_columns_r2018(
    content: str, columns: MTextColumns, dxfattribs
) -> MText:
    if not content:
        raise ValueError("no content")
        # column count is not required for DXF R2018
    mtext = MText.new(dxfattribs=dxfattribs)
    mtext.setup_columns(columns, linked=False)
    mtext.text = content
    return mtext



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License
import pytest

pytest.importorskip("PySide6")

from io import StringIO
import math
from ezdxf.lldxf.tags import Tags, DXFTag
from ezdxf.lldxf.loader import load_dxf_structure
from ezdxf.lldxf.tagger import ascii_tags_loader

from ezdxf.addons.browser import DXFTagsModel, DXFStructureModel, DXFDocument
from ezdxf.addons.browser.tags import compile_tags
from ezdxf.addons.browser.data import (
    EntityIndex,
    EntityHistory,
    SearchIndex,
)

from ezdxf.addons.xqt import Qt, QModelIndex


def txt2tags(s: str) -> Tags:
    return Tags(ascii_tags_loader(StringIO(s), skip_comments=False))


NAN = float("nan")


# noinspection PyMissingConstructor
class ModelIndex(QModelIndex):
    """Proxy"""

    def __init__(self, row, col):
        self._row = row
        self._col = col

    def row(self):
        return self._row

    def column(self):
        return self._col


class TestDXFTagsModel:
    def tags(self):
        return txt2tags(POINT)

    @pytest.fixture
    def model(self):
        return DXFTagsModel(self.tags())

    def test_fixed_column_count(self, model):
        assert model.columnCount() == 3

    def test_row_count(self, model):
        assert model.rowCount() == len(compile_tags(self.tags()))

    def test_render_display_role(self, model):
        assert model.data(ModelIndex(0, 0), role=Qt.DisplayRole) == "0"
        assert model.data(ModelIndex(0, 1), role=Qt.DisplayRole) == "<ctrl>"
        assert model.data(ModelIndex(0, 2), role=Qt.DisplayRole) == "POINT"


POINT = """0
POINT
5
0
330
0
100
AcDbEntity
8
0
100
AcDbPoint
10
0.0
20
0.0
30
0.0
"""


def test_setup_dxf_structure_model():
    sections = load_dxf_structure(txt2tags(ENTITIES))
    doc = DXFDocument(sections)
    model = DXFStructureModel("ez.dxf", doc)
    parent = model.item(0, 0)
    assert parent.data(Qt.DisplayRole) == "ez.dxf"
    assert "ENTITIES" in parent.child(0, 0).data(Qt.DisplayRole)
    # one level down
    parent = parent.child(0, 0)
    assert "LINE" in parent.child(0, 0).data(Qt.DisplayRole)
    assert "LINE" in parent.child(1, 0).data(Qt.DisplayRole)


class TestDXFDocument:
    @pytest.fixture
    def doc(self):
        sections = load_dxf_structure(txt2tags(ENTITIES))
        return DXFDocument(sections)

    def test_get_entity_returns_entity_tags(self, doc):
        entity = doc.get_entity("100")
        assert entity[0] == (0, "LINE")

    def test_get_entity_by_invalid_handle_returns_none(self, doc):
        assert doc.get_entity("XXX") is None

    def test_get_start_line_number_for_entity(self, doc):
        entity = doc.get_entity("101")
        assert doc.get_line_number(entity) == 9

    def test_get_entity_by_line_number(self, doc):
        entity = doc.get_entity("101")
        assert doc.get_entity_at_line(9) is entity
        assert doc.get_entity_at_line(10) is entity
        assert (
            doc.get_entity_at_line(99) is entity
        ), "should return the last entity"


class TestTagCompiler:
    def test_compile_single_int(self):
        tags = compile_tags(txt2tags("70\n3"))
        assert tags[0] == (70, 3)

    def test_compile_invalid_int_to_str(self):
        tags = compile_tags(txt2tags("70\nx"))
        assert tags[0] == (70, "x")

    def test_compile_single_float(self):
        tags = compile_tags(txt2tags("40\n3.14"))
        assert tags[0] == (40, 3.14)

    def test_compile_invalid_float_to_str(self):
        tags = compile_tags(txt2tags("40\nx.14"))
        assert tags[0] == (40, "x.14")

    def test_compile_single_2d_point(self):
        tags = compile_tags(txt2tags("10\n1.2\n20\n2.3"))
        assert tags[0] == (10, (1.2, 2.3))

    def test_compile_two_2d_points(self):
        tags = compile_tags(txt2tags("10\n1.1\n20\n1.2\n10\n2.1\n20\n2.2"))
        assert tags[0] == (10, (1.1, 1.2))
        assert tags[1] == (10, (2.1, 2.2))

    def test_compile_nan_coords_2d(self):
        tags = compile_tags(txt2tags("10\nx.2\n20\n2.3"))
        assert math.isnan(tags[0].value[0])

    def test_compile_single_3d_point(self):
        tags = compile_tags(txt2tags("10\n1.2\n20\n2.3\n30\n3.4"))
        assert tags[0] == (10, (1.2, 2.3, 3.4))

    def test_compile_nan_coords_3d(self):
        tags = compile_tags(txt2tags("10\n1\n20\n2\n30\nx"))
        assert math.isnan(tags[0].value[2])

    def test_compile_single_group_code_10(self):
        tags = compile_tags(txt2tags("10\n1.1"))
        assert tags[0] == (10, 1.1)

    def test_compile_two_group_code_10(self):
        tags = compile_tags(txt2tags("10\n1.1\n10\n2.2"))
        assert tags[0] == (10, 1.1)
        assert tags[1] == (10, 2.2)

    def test_compile_swapped_coords(self):
        tags = compile_tags(txt2tags("20\n2.2\n10\n1.1"))
        assert tags[0] == (20, 2.2), "expected coords as single tags"
        assert tags[1] == (10, 1.1), "expected coords as single tags"


ENTITIES = """0
SECTION
2
ENTITIES
0
LINE
5
100
0
LINE
5
101
0
ENDSEC
0
EOF
"""

SECTIONS = """0
SECTION
2
HEADER
9
$ACADVER
1
AC1032
0
ENDSEC
0
SECTION
2
ENTITIES
0
LINE
5
100
0
ENDSEC
0
EOF
"""


class TestEntityIndex:
    @pytest.fixture(scope="class")
    def index(self):
        data = {
            "ENTITIES": [
                Tags([DXFTag(0, "ENTITY1"), DXFTag(5, "F001")]),
                Tags([DXFTag(0, "ENTITY2"), DXFTag(5, "F002")]),
                Tags([DXFTag(0, "ENTITY3"), DXFTag(5, "F003")]),
                Tags([DXFTag(0, "ENTITY4"), DXFTag(5, "F004")]),
                # last entity without handle, has dummy handle "*1"
                Tags([DXFTag(0, "ENTITY5"), DXFTag(1, "DATA")]),
            ]
        }
        return EntityIndex(data)

    def test_contains_all_entities(self, index):
        assert "F001" in index
        assert "F002" in index
        assert "F003" in index
        assert "F004" in index
        assert "*1" in index, "expected dummy handle"

    def test_get_entity_by_handle(self, index):
        tags = index.get("F001")
        assert tags[0] == (0, "ENTITY1")

    def test_get_entity_by_dummy_handle(self, index):
        tags = index.get("*1")
        assert tags[0] == (0, "ENTITY5")

    def test_get_handle_from_casted_tags(self, index):
        entity = Tags(index.get("F001"))
        assert index.get_handle(entity) == "F001"

    def test_get_dummy_handle_from_casted_tags(self, index):
        entity = Tags(index.get("*1"))
        assert index.get_handle(entity) == "*1"

    def test_get_next_entity(self, index):
        e1 = index.get("F001")
        e2 = index.get("F002")
        assert index.next_entity(e1) is e2

    def test_next_entity_of_last_entity_is_last_entity(self, index):
        e1 = index.get("*1")
        assert index.next_entity(e1) is e1

    def test_get_prev_entity(self, index):
        e1 = index.get("F001")
        e2 = index.get("F002")
        assert index.previous_entity(e2) is e1

    def test_prev_entity_of_first_entity_is_first_entity(self, index):
        e1 = index.get("F001")
        assert index.previous_entity(e1) is e1

    def test_max_line_number(self, index):
        assert index.max_line_number == 20

    def test_get_start_line_number(self, index):
        e = index.get("F003")
        assert index.get_start_line_for_entity(e) == 9

    def test_get_start_line_number_for_dummy_handle(self, index):
        e = index.get("*1")
        assert index.get_start_line_for_entity(e) == 17

    def test_entity_at_line(self, index):
        e3 = index.get("F003")
        assert index.get_entity_at_line(9) is e3
        assert index.get_entity_at_line(10) is e3

    def test_entity_at_line_for_dummy_handle(self, index):
        e = index.get("*1")
        assert index.get_entity_at_line(19) is e
        assert index.get_entity_at_line(20) is e


def test_entity_index_adds_missing_endsec_tag():
    # The function load_dxf_structure() throws the ENDSEC tag away.
    # The entity indexer must take this issue into account!
    sections = load_dxf_structure(txt2tags(SECTIONS))
    index = EntityIndex(sections)
    entity = index.get_entity_at_line(15)
    assert entity.get_handle() == "100"
    assert index.get_start_line_for_entity(entity) == 15


class TestEntityHistory:
    @pytest.fixture
    def history2(self):
        history = EntityHistory()
        history.append(Tags([DXFTag(1, "first")]))
        history.append(Tags([DXFTag(2, "second")]))
        return history

    def test_setup_history(self):
        history = EntityHistory()
        assert len(history) == 0
        assert history.index == 0

    def test_empty_history_returns_none(self):
        history = EntityHistory()
        assert history.back() is None
        assert history.forward() is None

    def test_append_one_entity(self):
        history = EntityHistory()
        history.append(Tags())
        assert len(history) == 1
        assert history.index == 0

    def test_append_two_entities(self):
        history = EntityHistory()
        history.append(Tags())
        history.append(Tags())
        assert len(history) == 2
        assert history.index == 1

    def test_go_back_in_history(self, history2):
        first, second = history2.content()
        assert history2.index == 1
        assert history2.back() is first
        assert len(history2) == 2, "entity is still in history"
        assert history2.index == 0

    def test_go_back_and_forward_in_history(self, history2):
        first, second = history2.content()
        assert history2.back() is first
        assert history2.forward() is second

    def test_append_should_add_time_travel_history(self, history2):
        first, second = history2.content()
        assert history2.back() is first  # 1st time travel
        assert history2.index == 0

        assert history2.forward() is second  # 2nd time travel
        assert history2.index == 1

        third = Tags([DXFTag(3, "third")])
        history2.append(third)
        assert history2.index == 4

        # complete travel history
        content = history2.content()
        assert len(content) == 5
        #                                 time wraps ->  append
        assert content == [first, second, first, second, third]


SEARCH_EXAMPLE1 = """0
SEARCH1
8
LayerName1
62
7
"""

SEARCH_EXAMPLE2 = """0
SEARCH2
8
LayerName2
62
6
"""


class TestSearchIndex:
    @pytest.fixture(scope="class")
    def entities(self):
        return [txt2tags(SEARCH_EXAMPLE1), txt2tags(SEARCH_EXAMPLE2)]

    @pytest.fixture
    def search(self, entities):
        return SearchIndex(entities)

    @staticmethod
    def move_cursor_forward(s: SearchIndex, count: int):
        for _ in range(count):
            s.move_cursor_forward()

    @staticmethod
    def move_cursor_backward(s: SearchIndex, count: int):
        for _ in range(count):
            s.move_cursor_backward()

    def test_valid_setup_and_default_settings(self, search):
        assert len(search.entities) == 2
        assert search.is_end_of_index is False
        assert (
            search.case_insensitive is True
        ), "should be case insensitive by default"
        assert (
            search.numbers is False
        ), "should not search in number tags by default"

    def test_empty_search_index(self):
        search_index = SearchIndex([])
        assert search_index.is_end_of_index is True

    def test_reset_cursor_forward(self, search):
        search.reset_cursor(backward=False)
        assert search.cursor() == (
            0,
            0,
        ), "cursor should be the first tag of the first entity"
        assert search.is_end_of_index is False

    def test_move_cursor_forward(self, search):
        search.reset_cursor()
        search.move_cursor_forward()
        assert search.cursor() == (0, 1)

    def test_move_cursor_forward_beyond_entity_border(self, search):
        search.reset_cursor()
        self.move_cursor_forward(search, 3)
        assert search.cursor() == (1, 0)

    def test_move_cursor_forward_to_the_end_of_index(self, search):
        search.reset_cursor()
        self.move_cursor_forward(search, 10)
        assert search.is_end_of_index is True
        assert search.cursor() == (
            1,
            2,
        ), "index should stop at the last tag of the last entity"

    def test_reset_cursor_backward(self, search):
        search.reset_cursor(backward=True)
        assert search.cursor() == (
            1,
            2,
        ), "cursor should be the last tag of the last entity"
        assert search.is_end_of_index is False

    def test_move_cursor_backward(self, search):
        search.reset_cursor(backward=True)
        search.move_cursor_backward()
        assert search.cursor() == (1, 1)

    def test_move_cursor_backward_beyond_entity_border(self, search):
        search.reset_cursor(backward=True)
        self.move_cursor_backward(search, 3)
        assert search.cursor() == (0, 2)

    def test_move_cursor_backward_to_the_end_of_index(self, search):
        search.reset_cursor()
        self.move_cursor_backward(search, 10)
        assert search.is_end_of_index is True
        assert search.cursor() == (
            0,
            0,
        ), "index should stop at the first tag of the first entity"

    def test_failing_search(self, search):
        entity, index = search.find("XDATA")
        assert entity is None
        assert index == -1
        assert search.is_end_of_index is True

    def test_find_entity_type(self, search):
        entity, index = search.find("SEARCH1")
        assert entity is search.entities[0]
        assert index == 0

    def test_find_forward_entity_type(self, search):
        search.find("SEARCH")
        entity, index = search.find_forward()
        assert entity is search.entities[1]
        assert index == 0

    def test_find_content(self, search):
        entity, index = search.find("LayerName1")
        assert entity is search.entities[0]
        assert index == 1

    def test_find_forward_content(self, search):
        search.find("LayerName")
        entity, index = search.find_forward()
        assert entity is search.entities[1]
        assert index == 1

    def test_failing_find_forward_returns_none(self, search):
        search.find("LayerName")
        search.find_forward()
        entity, index = search.find_forward()
        assert entity is None
        assert index == -1

    def test_not_initiated_find_forward_returns_none(self, search):
        entity, index = search.find_forward()
        assert entity is None
        assert index == -1

    def test_case_insensitive_search(self, search):
        search.case_insensitive = True
        entity, index = search.find("LAYERNAME1")
        assert entity is search.entities[0]
        assert index == 1

    def test_case_sensitive_search(self, search):
        search.case_insensitive = False
        entity, index = search.find("LAYERNAME1")
        assert entity is None

    def test_ignore_number_tags(self, search):
        search.numbers = False
        entity, index = search.find("6")
        assert entity is None

    def test_search_in_number_tags(self, search):
        search.numbers = True
        entity, index = search.find("6")
        assert entity is search.entities[1]
        assert index == 2

    def test_failing_find_forward_stops_at_the_end(self, search):
        assert search.find("XXX") is search.NOT_FOUND
        assert search.is_end_of_index is True

    def test_failing_find_backwards_stops_at_the_beginning(self, search):
        assert search.find("XXX", backward=True) is search.NOT_FOUND
        assert search.is_end_of_index is True


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2020 Manfred Moitzi
# License: MIT License
import pytest
import math
import ezdxf
from ezdxf.layouts import VirtualLayout

POINTS = [(0, 0), (1, 0, 1), (2, 0), (3, 0)]


@pytest.fixture(scope="module")
def msp():
    doc = ezdxf.new()
    return doc.modelspace()


@pytest.fixture
def polyline2d(msp):
    return msp.add_polyline2d(
        points=POINTS, format="xyb", dxfattribs={"layer": "LAY", "color": 1}
    )


def test_polyline2d_virtual_entities(polyline2d):
    result = list(polyline2d.virtual_entities())
    assert len(result) == 3

    e = result[0]
    assert e.dxftype() == "LINE"
    assert e.dxf.layer == "LAY"
    assert e.dxf.color == 1
    assert e.dxf.start == (0, 0)
    assert e.dxf.end == (1, 0)

    e = result[1]
    assert e.dxftype() == "ARC"
    assert e.dxf.layer == "LAY"
    assert e.dxf.color == 1
    assert e.dxf.center == (1.5, 0)
    assert e.dxf.radius == 0.5
    assert math.isclose(e.dxf.start_angle, 180, abs_tol=1e-12)
    assert math.isclose(e.dxf.end_angle, 0, abs_tol=1e-12)

    assert e.start_point.isclose((1, 0))
    assert e.end_point.isclose((2, 0))

    e = result[2]
    assert e.dxftype() == "LINE"
    assert e.dxf.layer == "LAY"
    assert e.dxf.color == 1
    assert e.dxf.start == (2, 0)
    assert e.dxf.end == (3, 0)


def test_virtual_sub_entities_source_tracking(polyline2d):
    result = set(e.source_of_copy for e in polyline2d.virtual_entities())
    assert len(result) == 1, "only one source of copy expected"
    assert polyline2d in result, "polyline2d should be the source of copy"


def test_polyline2d_elevation(polyline2d):
    polyline = polyline2d.translate(1, 1, 1)
    assert polyline.dxf.elevation == (0, 0, 1)
    result = list(polyline.virtual_entities())
    assert len(result) == 3
    e = result[0]
    assert e.dxftype() == "LINE"
    assert e.dxf.start == (1, 1, 1)
    assert e.dxf.end == (2, 1, 1)

    e = result[1]
    assert e.dxftype() == "ARC"
    assert e.dxf.center == (2.5, 1, 1)
    assert e.dxf.radius == 0.5
    assert math.isclose(e.dxf.start_angle, 180, abs_tol=1e-12)
    assert math.isclose(e.dxf.end_angle, 0, abs_tol=1e-12)

    assert e.start_point.isclose((2, 1, 1))
    assert e.end_point.isclose((3, 1, 1))

    e = result[2]
    assert e.dxftype() == "LINE"
    assert e.dxf.start == (3, 1, 1)
    assert e.dxf.end == (4, 1, 1)


def test_polyline2d_closed():
    # Create a circle by 2D POLYLINE:
    msp = VirtualLayout()
    polyline = msp.add_polyline2d(points=[(0, 0, 1), (1, 0, 1)], format="xyb")
    polyline.close(True)

    result = list(polyline.virtual_entities())
    assert len(result) == 2

    e = result[0]
    assert e.dxftype() == "ARC"
    assert e.dxf.center == (0.5, 0)
    assert e.dxf.radius == 0.5
    assert math.isclose(e.dxf.start_angle, 180, abs_tol=1e-12)
    assert math.isclose(e.dxf.end_angle, 0, abs_tol=1e-12)

    e = result[1]
    assert e.dxftype() == "ARC"
    assert e.dxf.center.isclose((0.5, 0))
    assert e.dxf.radius == 0.5
    assert math.isclose(e.dxf.start_angle, 0, abs_tol=1e-12)
    assert math.isclose(abs(e.dxf.end_angle), 180, abs_tol=1e-12)


def test_polyline2d_explode(msp):
    # explode does not work with VirtualLayout()
    polyline = msp.add_polyline2d(POINTS, format="xyb")
    count = len(msp)
    result = polyline.explode()
    assert polyline.is_alive is False
    assert len(msp) == count + 2  # LINE, ARC, LINE
    assert msp[-1] is result[2]
    assert msp[-2] is result[1]
    assert msp[-3] is result[0]


def test_polyline3d_virtual_entities():
    msp = VirtualLayout()
    polyline3d = msp.add_polyline3d([(0, 0, 0), (1, 0, 0), (2, 2, 2)])
    result = list(polyline3d.virtual_entities())
    assert len(result) == 2
    line = result[0]
    assert line.dxftype() == "LINE"
    assert line.dxf.start == (0, 0, 0)
    assert line.dxf.end == (1, 0, 0)
    line = result[1]
    assert line.dxftype() == "LINE"
    assert line.dxf.start == (1, 0, 0)
    assert line.dxf.end == (2, 2, 2)


def test_polyline3d_closed():
    msp = VirtualLayout()
    polyline3d = msp.add_polyline3d(
        [(0, 0, 0), (1, 0, 0), (2, 2, 2)], close=True
    )
    assert polyline3d.is_closed is True
    result = list(polyline3d.virtual_entities())
    assert len(result) == 3
    # The closing element is the first LINE entity.
    # This is an implementation detail and can change in the future!
    line = result[0]
    assert line.dxftype() == "LINE"
    assert line.dxf.start == (2, 2, 2)
    assert line.dxf.end == (0, 0, 0)


def test_polyline3d_explode(msp):
    # explode does not work with VirtualLayout()
    polyline3d = msp.add_polyline3d([(0, 0, 0), (1, 0, 0), (2, 2, 2)])
    count = len(msp)
    result = polyline3d.explode()
    assert polyline3d.is_alive is False
    assert len(msp) == count + 1  # LINE, LINE
    assert msp[-1] is result[1]
    assert msp[-2] is result[0]


@pytest.fixture()
def polymesh():
    msp = VirtualLayout()
    polymesh = msp.add_polymesh((3, 3))
    for m in range(3):
        for n in range(3):
            polymesh.set_mesh_vertex((m, n), (m, n))
    return polymesh


def test_polymesh_virtual_entities(polymesh):
    result = list(polymesh.virtual_entities())
    assert len(result) == 4
    assert result[0].dxftype() == "3DFACE"
    # 1. columns, 2. rows
    # col=0, row=0
    assert result[0].dxf.vtx0 == (0, 0, 0)
    assert result[0].dxf.vtx2 == (1, 1, 0)
    # col=0, row=1
    assert result[1].dxf.vtx0 == (0, 1, 0)
    assert result[1].dxf.vtx2 == (1, 2, 0)
    # col=1, row=0
    assert result[2].dxf.vtx0 == (1, 0, 0)
    assert result[2].dxf.vtx2 == (2, 1, 0)
    # col=1, row=1
    assert result[3].dxf.vtx0 == (1, 1, 0)
    assert result[3].dxf.vtx2 == (2, 2, 0)


def test_closed_polymesh(polymesh):
    polymesh.close(m_close=True, n_close=True)
    assert polymesh.is_m_closed is True
    assert polymesh.is_n_closed is True
    result = list(polymesh.virtual_entities())
    assert len(result) == 9


def test_polyface_virtual_entities():
    from ezdxf.render.forms import cube

    msp = VirtualLayout()
    polyface = cube().render_polyface(msp)
    result = list(polyface.virtual_entities())

    assert len(result) == 6
    assert result[0].dxftype() == "3DFACE"
    vertices = set()
    for face in result:
        for vertex in face:
            vertices.add(vertex)
    assert len(vertices) == 8
    assert (-0.5, -0.5, -0.5) in vertices
    assert (0.5, 0.5, 0.5) in vertices



# Copyright (c) 2011-2019, Manfred Moitzi
# License: MIT License
import pytest

import ezdxf
from ezdxf.tools.test import load_entities
from ezdxf.sections.blocks import BlocksSection
from ezdxf.lldxf.tagwriter import TagCollector
from ezdxf.entities import factory
from ezdxf.lldxf.const import BLK_NON_CONSTANT_ATTRIBUTES


@pytest.fixture
def dxf12():
    return ezdxf.new("R12")


@pytest.fixture
def blocks(dxf12):
    return BlocksSection(dxf12, list(load_entities(TESTBLOCKS, "BLOCKS")))


@pytest.fixture
def bounded_blocks(dxf12):
    entities = list(load_entities(TESTBLOCKS, "BLOCKS"))
    for entity in entities:
        factory.bind(entity, dxf12)
    return BlocksSection(dxf12, entities)


@pytest.fixture
def doc():
    doc = ezdxf.new()
    doc.blocks.new("_ARCHTICK")
    doc.blocks.new("_OPEN30")
    return doc


def test_empty_section(dxf12):
    blocks = BlocksSection(dxf12, list(load_entities(EMPTYSEC, "BLOCKS")))
    # the NES creates automatically *Model_Space and *Paper_Space blocks
    assert "*Model_Space" in blocks
    assert "*Paper_Space" in blocks

    collector = TagCollector(dxfversion=dxf12.dxfversion)
    blocks.export_dxf(collector)

    assert collector.tags[0] == (0, "SECTION")
    assert collector.tags[1] == (2, "BLOCKS")
    assert collector.tags[2] == (0, "BLOCK")
    # tag[3] is a arbitrary handle
    assert collector.tags[3][0] == 5
    assert collector.tags[4] == (8, "0")  # default layer '0'
    assert collector.tags[5] == (
        2,
        "$Model_Space",
    )  # export modelspace with leading '$' for R12
    assert collector.tags[-1] == (0, "ENDSEC")


def test_key(blocks):
    assert blocks.key("Test") == "test"
    block = blocks.new("TEST")
    assert blocks.key(block) == "test"


def test_is_layout_block(blocks):
    block = blocks.new("TEST")
    assert block.is_any_layout is False

    # required modelspace block already created
    msp = blocks.get("*Model_Space")
    assert msp.is_modelspace is True

    # required paperspace block already created
    psp = blocks.get("*Paper_Space")
    assert psp.is_any_paperspace is True
    assert psp.is_active_paperspace is True


def test_overwrite_existing_block(blocks):
    block = blocks.new("TEST")
    assert block.dxf.name in blocks
    old_len = len(blocks)
    with pytest.raises(ezdxf.DXFTableEntryError):
        # can not create block with existing name
        blocks.new("Test")  # block names are case insensitive
    assert len(blocks) == old_len, 'should not create block "TEST"'

    blocks.delete_block("Test", safe=False)
    assert len(blocks) == old_len - 1, 'should remove existing block "TEST"'
    blocks.new("Test")
    assert len(blocks) == old_len, 'should create new block "Test"'


def test_not_in_blocks_section(blocks):
    assert "TEST" not in blocks


def test_getitem(blocks):
    blocks.new("TEST")
    block = blocks["TEST"]
    assert "TEST" == block.name
    block = blocks["Test"]
    assert "TEST" == block.name


def test_new_block_layout(doc):
    block = doc.blocks.new("NewBlockLayout")
    block.add_point((0, 0, 0))
    assert len(block) == 1
    assert block.can_explode is True
    assert block.scale_uniformly is False

    block.can_explode = False
    block.scale_uniformly = True

    assert block.block_record.dxf.explode == 0
    assert block.block_record.dxf.scale == 1


def test_case_insensitivity(blocks):
    blocks.new("TEST")
    assert "TEST" in blocks
    assert "Test" in blocks


def test_iter_blocks(blocks):
    blocks = list(blocks)
    assert 4 == len(blocks)


def test_block_content_entity_drawing_attribute(blocks, dxf12):
    archtick = blocks["_ARCHTICK"]
    entities = list(archtick)
    assert 1 == len(entities)  # VERTEX & SEQEND doesn't count


def test_delete_block(bounded_blocks, dxf12):
    archtick = bounded_blocks["_ARCHTICK"]
    entities = list(archtick)
    archtick_name = archtick.name
    bounded_blocks.delete_block(archtick_name, safe=False)
    assert archtick_name not in bounded_blocks
    assert archtick.is_alive is False
    for entity in entities:
        assert entity.is_alive is False


@pytest.mark.parametrize(
    "name",
    [
        "*Model_Space",
        "*Paper_Space",
        "_ARCHTICK",
        "_OPEN30",
    ],
)
def test_do_not_delete_layouts_and_special_arrow_blocks_in_safe_mode(doc, name):
    with pytest.raises(ezdxf.DXFBlockInUseError):
        doc.blocks.delete_block(name, safe=True)


def test_delete_all_except_layouts_and_special_arrow_blocks(doc):
    doc.blocks.delete_all_blocks()
    assert len(doc.blocks) == 4
    block_names = set(block.name for block in doc.blocks)
    assert block_names == {
        "*Model_Space",
        "*Paper_Space",
        "_ARCHTICK",
        "_OPEN30",
    }


def test_rename_block(blocks):
    block = blocks.new("RENAME_ME")
    assert block.dxf.name in blocks

    blocks.rename_block("RENAME_ME", "NEW_NAME")
    assert "NEW_NAME" in blocks

    # block names are case insensitive
    blocks.rename_block("New_Name", "check_lower_case")
    assert "Check_Lower_Case" in blocks

    # but originals name is preserved
    assert blocks["Check_Lower_Case"].name == "check_lower_case"


@pytest.fixture(scope="module")
def dxf2000():
    return ezdxf.new("R2000")


@pytest.fixture
def dxf2000_blocks(dxf2000):
    if "TestBlock" not in dxf2000.blocks:
        block = dxf2000.blocks.new("TestBlock")
        block.add_line((0, 0), (10, 10))
        block.add_line((0, 0), (10, 10))
        block.add_line((0, 0), (10, 10))
    return dxf2000.blocks


def test_dxf2000_dxf_block_structure(dxf2000_blocks, dxf2000):
    assert "TestBlock" in dxf2000_blocks
    block = dxf2000_blocks["TestBlock"]
    block_record_handle = block.block_record_handle

    # exists an associated block record entry?
    block_record = dxf2000.tables.block_records.get(block.name)
    assert block_record_handle == block_record.dxf.handle
    assert block_record.dxf.name == block.name


def test_dxf2000_delete_block(dxf2000_blocks, dxf2000):
    block = dxf2000_blocks["TestBlock"]
    block_name = block.name
    entities = list(block)
    block_record_handle = block.block_record_handle

    block_count = len(dxf2000_blocks)
    dxf2000_blocks.delete_block(block_name)

    # removed from blocks load_section?
    assert block_count - 1 == len(dxf2000_blocks)
    assert block_name not in dxf2000_blocks

    # all block related management data deleted?
    assert block.is_alive is False

    # removed from block records table?
    assert block_name not in dxf2000.tables.block_records

    # all entities deleted ?
    for entity in entities:
        assert entity.is_alive is False
    # we are done!


def test_dxf2000_delete_all_blocks(dxf2000_blocks):
    dxf2000_blocks.delete_all_blocks()
    blocks = list(dxf2000_blocks)

    # assure not deleting layout blocks or arrow blocks
    assert len(blocks) == 2

    block_names = [block.name for block in blocks]
    block_names.sort()
    assert ["*Model_Space", "*Paper_Space"] == block_names


def test_dxf2000_rename_block(dxf2000_blocks):
    block = dxf2000_blocks.new("RENAME_ME")
    assert block.dxf.name in dxf2000_blocks

    dxf2000_blocks.rename_block("RENAME_ME", "NEW_NAME")
    assert "NEW_NAME" in dxf2000_blocks


def test_update_block_flags(doc):
    blk = doc.blocks.new("UPDATE_BLOCK_FLAGS")
    blk.add_attdef("TEST", (0, 0))
    assert blk.block.get_flag_state(BLK_NON_CONSTANT_ATTRIBUTES) is False
    blk.update_block_flags()
    assert blk.block.get_flag_state(BLK_NON_CONSTANT_ATTRIBUTES) is True


EMPTYSEC = """  0
SECTION
  2
BLOCKS
  0
ENDSEC
"""

TESTBLOCKS = """  0
SECTION
  2
BLOCKS
  0
BLOCK
  8
0
  2
$MODEL_SPACE
 70
     0
 10
0.0
 20
0.0
 30
0.0
  3
$MODEL_SPACE
  1

  0
ENDBLK
  5
21
  8
0
  0
BLOCK
 67
     1
  8
0
  2
$PAPER_SPACE
 70
     0
 10
0.0
 20
0.0
 30
0.0
  3
$PAPER_SPACE
  1

  0
ENDBLK
  5
5B
 67
     1
  8
0
  0
BLOCK
  8
0
  2
_ARCHTICK
 70
     0
 10
0.0
 20
0.0
 30
0.0
  3
_ARCHTICK
  1

  0
POLYLINE
  5
239
  8
0
  6
BYBLOCK
 62
     0
 66
     1
 10
0.0
 20
0.0
 30
0.0
 40
0.15
 41
0.15
  0
VERTEX
  5
403
  8
0
  6
BYBLOCK
 62
     0
 10
-0.5
 20
-0.5
 30
0.0
  0
VERTEX
  5
404
  8
0
  6
BYBLOCK
 62
     0
 10
0.5
 20
0.5
 30
0.0
  0
SEQEND
  5
405
  8
0
  6
BYBLOCK
 62
     0
  0
ENDBLK
  5
23B
  8
0
  0
BLOCK
  8
0
  2
_OPEN30
 70
     0
 10
0.0
 20
0.0
 30
0.0
  3
_OPEN30
  1

  0
LINE
  5
23D
  8
0
  6
BYBLOCK
 62
     0
 10
-1.0
 20
0.26794919
 30
0.0
 11
0.0
 21
0.0
 31
0.0
  0
LINE
  5
23E
  8
0
  6
BYBLOCK
 62
     0
 10
0.0
 20
0.0
 30
0.0
 11
-1.0
 21
-0.26794919
 31
0.0
  0
LINE
  5
23F
  8
0
  6
BYBLOCK
 62
     0
 10
0.0
 20
0.0
 30
0.0
 11
-1.0
 21
0.0
 31
0.0
  0
ENDBLK
  5
241
  8
0
  0
ENDSEC
"""



#  Copyright (c) 2023-2024, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import (
    Iterable,
    Iterator,
    NamedTuple,
    Optional,
    Tuple,
    List,
    TYPE_CHECKING,
    Sequence,
)
import enum
import logging

from ezdxf.math import (
    Matrix44,
    UVec,
    Vec3,
    NonUniformScalingError,
    InsertTransformationError,
)
from ezdxf.entities import (
    DXFEntity,
    DXFGraphic,
    Circle,
    LWPolyline,
    Polyline,
    Ellipse,
    is_graphic_entity,
)
from ezdxf.entities.copy import default_copy, CopyNotSupported
from ezdxf.protocols import SupportsTemporaryTransformation
from ezdxf.document import Drawing

if TYPE_CHECKING:
    from ezdxf.layouts import BlockLayout

__all__ = [
    "Logger",
    "Error",
    "inplace",
    "copies",
    "translate",
    "scale_uniform",
    "scale",
    "x_rotate",
    "y_rotate",
    "z_rotate",
    "axis_rotate",
    "transform_entity_by_blockref",
    "transform_entities_by_blockref",
]
MIN_SCALING_FACTOR = 1e-12
logger = logging.getLogger("ezdxf")


class Error(enum.Enum):
    NONE = 0
    TRANSFORMATION_NOT_SUPPORTED = enum.auto()
    COPY_NOT_SUPPORTED = enum.auto()
    NON_UNIFORM_SCALING_ERROR = enum.auto()
    INSERT_TRANSFORMATION_ERROR = enum.auto()
    VIRTUAL_ENTITY_NOT_SUPPORTED = enum.auto()


class Logger:
    class Entry(NamedTuple):
        error: Error
        message: str
        entity: DXFEntity

    def __init__(self) -> None:
        self._entries: list[Logger.Entry] = []

    def __getitem__(self, index: int) -> Entry:
        """Returns the error entry at `index`."""
        return self._entries[index]

    def __iter__(self) -> Iterator[Entry]:
        """Iterates over all error entries."""
        return iter(self._entries)

    def __len__(self) -> int:
        """Returns the count of error entries."""
        return len(self._entries)

    def add(self, error: Error, message: str, entity: DXFEntity):
        self._entries.append(Logger.Entry(error, message, entity))

    def messages(self) -> list[str]:
        """Returns all error messages as list of strings."""
        return [entry.message for entry in self._entries]

    def purge(self, entries: Iterable[Entry]) -> None:
        for entry in entries:
            try:
                self._entries.remove(entry)
            except ValueError:
                pass


def _inplace(entities: Iterable[DXFEntity], m: Matrix44) -> Logger:
    """Transforms the given `entities` inplace by the transformation matrix `m`,
    non-uniform scaling is not supported. The function logs errors and does not raise
    errors for unsupported entities or transformations that cannot be performed,
    see enum :class:`Error`.
    The :func:`inplace` function supports virtual entities as well.

    """
    log = Logger()
    for entity in entities:
        if not entity.is_alive:
            continue
        try:
            entity.transform(m)  # type: ignore
        except (AttributeError, NotImplementedError):
            log.add(
                Error.TRANSFORMATION_NOT_SUPPORTED,
                f"{str(entity)} entity does not support transformation",
                entity,
            )
        except NonUniformScalingError:
            log.add(
                Error.NON_UNIFORM_SCALING_ERROR,
                f"{str(entity)} entity does not support non-uniform scaling",
                entity,
            )
        except InsertTransformationError:
            log.add(
                Error.INSERT_TRANSFORMATION_ERROR,
                f"{str(entity)} entity can not represent a non-orthogonal target coordinate system",
                entity,
            )

    return log


def inplace(
    entities: Iterable[DXFEntity],
    m: Matrix44,
) -> Logger:
    """Transforms the given `entities` inplace by the transformation matrix `m`,
    non-uniform scaling is supported. The function converts circular arcs into ellipses
    to perform non-uniform scaling.  The function logs errors and does not raise errors
    for unsupported entities or transformation errors, see enum :class:`Error`.

    .. important::

        The :func:`inplace` function does not support type conversion for virtual
        entities e.g. non-uniform scaling for CIRCLE, ARC or POLYLINE with bulges,
        see also function :func:`copies`.

    """
    log = _inplace(entities, m)
    errors: list[Logger.Entry] = []
    for entry in log:
        if entry.error != Error.NON_UNIFORM_SCALING_ERROR:
            continue

        entity = entry.entity
        if entity.is_virtual:
            errors.append(entry)
            log.add(
                Error.VIRTUAL_ENTITY_NOT_SUPPORTED,
                f"non-uniform scaling is not supported for virtual entity {str(entity)}",
                entity,
            )
            continue

        if isinstance(entity, Circle):  # CIRCLE, ARC
            errors.append(entry)
            ellipse = entity.to_ellipse(replace=True)
            ellipse.transform(m)
        elif isinstance(entity, (LWPolyline, Polyline)):  # has bulges (circular arcs)
            errors.append(entry)
            for sub_entity in entity.explode():
                if isinstance(sub_entity, Circle):
                    sub_entity = sub_entity.to_ellipse()
                sub_entity.transform(m)  # type: ignore
        # else: NON_UNIFORM_SCALING_ERROR stays unchanged
    log.purge(errors)
    return log


def copies(
    entities: Iterable[DXFEntity], m: Optional[Matrix44] = None
) -> Tuple[Logger, List[DXFEntity]]:
    """Copy entities and transform them by matrix `m`. Does not raise any exception
    and ignores all entities that cannot be copied or transformed. Just copies the input
    entities if matrix `m` is ``None``. Returns a tuple of :class:`Logger` and a list of
    transformed virtual copies. The function supports virtual entities as input and
    converts circular arcs into ellipses to perform non-uniform scaling.
    """

    log = Logger()
    clones = _copy_entities(entities, log)
    if isinstance(m, Matrix44):
        clones = _transform_clones(clones, m, log)
    return log, clones


def _copy_entities(entities: Iterable[DXFEntity], log: Logger) -> list[DXFEntity]:
    clones: list[DXFEntity] = []
    for entity in entities:
        if not entity.is_alive:
            continue
        try:
            clone = entity.copy(copy_strategy=default_copy)
        except CopyNotSupported:
            log.add(
                Error.COPY_NOT_SUPPORTED,
                f"{str(entity)} entity does not support copy",
                entity,
            )
        else:
            clones.append(clone)
    return clones


def _transform_clones(clones: Iterable[DXFEntity], m: Matrix44, log: Logger):
    entities: List[DXFEntity] = []
    for entity in clones:
        try:
            entity.transform(m)  # type: ignore
        except (AttributeError, NotImplementedError):
            log.add(
                Error.TRANSFORMATION_NOT_SUPPORTED,
                f"{str(entity)} entity does not support transformation",
                entity,
            )
        except InsertTransformationError:
            log.add(
                Error.INSERT_TRANSFORMATION_ERROR,
                f"{str(entity)} entity can not represent a non-orthogonal target coordinate system",
                entity,
            )
        except NonUniformScalingError:
            try:
                entities.extend(_scale_non_uniform(entity, m))
            except TypeError:
                log.add(
                    Error.NON_UNIFORM_SCALING_ERROR,
                    f"{str(entity)} entity does not support non-uniform scaling",
                    entity,
                )
        else:
            entities.append(entity)

    return entities


def _scale_non_uniform(entity: DXFEntity, m: Matrix44):
    sub_entity: DXFGraphic
    if isinstance(entity, Circle):
        sub_entity = Ellipse.from_arc(entity)
        sub_entity.transform(m)
        yield sub_entity
    elif isinstance(entity, (LWPolyline, Polyline)):
        for sub_entity in entity.virtual_entities():
            if isinstance(sub_entity, Circle):
                sub_entity = Ellipse.from_arc(sub_entity)
            sub_entity.transform(m)
            yield sub_entity
    else:
        raise TypeError


def translate(entities: Iterable[DXFEntity], offset: UVec) -> Logger:
    """Translates (moves) `entities` inplace by the `offset` vector."""
    v = Vec3(offset)
    if v:
        return _inplace(entities, m=Matrix44.translate(v.x, v.y, v.z))
    return Logger()


def scale_uniform(entities: Iterable[DXFEntity], factor: float) -> Logger:
    """Scales `entities` inplace by a `factor` in all axis. Scaling factors smaller than
    :attr:`MIN_SCALING_FACTOR` are ignored.

    """
    f = float(factor)
    if abs(f) > MIN_SCALING_FACTOR:
        return _inplace(entities, m=Matrix44.scale(f, f, f))
    return Logger()


def scale(entities: Iterable[DXFEntity], sx: float, sy: float, sz: float) -> Logger:
    """Scales `entities` inplace by the factors `sx` in x-axis, `sy` in y-axis and `sz`
    in z-axis. Scaling factors smaller than :attr:`MIN_SCALING_FACTOR` are ignored.

    .. important::

        same limitations for virtual entities as the :func:`inplace` function

    """

    def safe(f: float) -> float:
        f = float(f)
        return f if abs(f) > MIN_SCALING_FACTOR else 1.0

    return inplace(entities, Matrix44.scale(safe(sx), safe(sy), safe(sz)))


def x_rotate(entities: Iterable[DXFEntity], angle: float) -> Logger:
    """Rotates `entities` inplace by `angle` in radians about the x-axis."""
    a = float(angle)
    if a:
        return _inplace(entities, m=Matrix44.x_rotate(a))
    return Logger()


def y_rotate(entities: Iterable[DXFEntity], angle: float) -> Logger:
    """Rotates `entities` inplace by `angle` in radians about the y-axis."""
    a = float(angle)
    if a:
        return _inplace(entities, m=Matrix44.y_rotate(a))
    return Logger()


def z_rotate(entities: Iterable[DXFEntity], angle: float) -> Logger:
    """Rotates `entities` inplace by `angle` in radians about the x-axis."""
    a = float(angle)
    if a:
        return _inplace(entities, m=Matrix44.z_rotate(a))
    return Logger()


def axis_rotate(entities: Iterable[DXFEntity], axis: UVec, angle: float) -> Logger:
    """Rotates `entities` inplace by `angle` in radians about the rotation axis starting
    at the origin pointing in `axis` direction.
    """
    a = float(angle)
    if not a:
        return Logger()

    v = Vec3(axis)
    if not v.is_null:
        return _inplace(entities, m=Matrix44.axis_rotate(v, a))
    return Logger()


def transform_entity_by_blockref(entity: DXFEntity, m: Matrix44) -> bool:
    """Apply a transformation by moving an entity into a block and replacing the entity
    by a block reference with the applied transformation.
    """
    return _transform_by_blockref([entity], m) is not None


def transform_entities_by_blockref(
    entities: Iterable[DXFEntity], m: Matrix44
) -> BlockLayout | None:
    """Apply a transformation by moving entities into a block and replacing the entities
    by a block reference with the applied transformation.
    """
    return _transform_by_blockref(list(entities), m)


def _transform_by_blockref(
    entities: Sequence[DXFEntity], m: Matrix44
) -> BlockLayout | None:
    if len(entities) == 0:
        return None

    first_entity = entities[0]
    if not is_graphic_entity(first_entity):
        return None

    doc = first_entity.doc
    if doc is None:
        return None

    layout = first_entity.get_layout()
    if layout is None:
        return None

    block = doc.blocks.new_anonymous_block()
    insert = layout.add_blockref(block.name, (0, 0, 0))
    try:
        insert.transform(m)
    except InsertTransformationError:
        logger.warning(f"cannot apply invalid transformation")
        layout.delete_entity(insert)
        doc.blocks.delete_block(block.name, safe=False)
        return None

    for e in entities:
        if is_graphic_entity(e):
            layout.move_to_layout(e, block)
    return block


def apply_temporary_transformations(entities: Iterable[DXFEntity]) -> None:
    for entity in entities:
        if isinstance(entity, SupportsTemporaryTransformation):
            tt = entity.temporary_transformation()
            tt.apply_transformation(entity)



# Copyright (c) 2020, Manfred Moitzi
# License: MIT License

import pytest

from ezdxf.math import has_clockwise_orientation


@pytest.fixture
def v1():
    return [(0, 0), (1, 0), (1, 1), (0, 1)]


@pytest.mark.parametrize(
    "vertices",
    [
        [(0, 0), (1, 0), (1, 1)],
        [(0, 0), (1, 0), (1, 1), (0, 1)],
    ],
)
def test_has_clockwise_orientation(vertices):
    assert has_clockwise_orientation(vertices) is False


@pytest.mark.parametrize(
    "vertices",
    [
        [(1, 1), (1, 0), (0, 0)],
        [(0, 1), (1, 1), (1, 0), (0, 0)],
    ],
)
def test_has_counter_clockwise_orientation(vertices):
    assert has_clockwise_orientation(vertices) is True


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2022, Manfred Moitzi
#  License: MIT License

import pytest
import ezdxf

from ezdxf.layouts import Paperspace
from ezdxf.entities.layer import LayerOverrides
from ezdxf.lldxf import const


@pytest.fixture(scope="module")
def doc():
    doc_ = ezdxf.new(setup=["linetypes"])
    doc_.layers.add("LayerA", color=ezdxf.colors.RED)
    doc_.layers.add("LayerB", color=ezdxf.colors.YELLOW)
    return doc_


@pytest.fixture
def layer_a(doc):
    return doc.layers.get("LayerA")


def test_doc_setup(doc):
    assert "LayerA" in doc.layers


def test_get_new_vp_override_object(layer_a):
    vp_overrides = layer_a.get_vp_overrides()
    assert isinstance(vp_overrides, LayerOverrides) is True


def test_if_a_layer_has_any_overrides(layer_a):
    vp_overrides = layer_a.get_vp_overrides()
    assert vp_overrides.has_overrides() is False


def test_if_a_layer_has_overrides_for_specific_viewport(layer_a):
    vp_overrides = layer_a.get_vp_overrides()
    assert vp_overrides.has_overrides("ABBA") is False


def test_get_default_layer_values_for_missing_overrides(layer_a):
    vp_handle = "ABBA"
    overrides = layer_a.get_vp_overrides()
    assert overrides.has_overrides(vp_handle) is False
    assert overrides.get_color(vp_handle) == ezdxf.colors.RED
    assert overrides.get_rgb(vp_handle) is None
    assert overrides.get_transparency(vp_handle) == 0.0
    assert overrides.get_linetype(vp_handle) == "Continuous"
    assert overrides.get_lineweight(vp_handle) == const.LINEWEIGHT_DEFAULT


class TestSetOverridesWithoutCommit:
    def test_set_color_override(self, layer_a):
        vp_overrides = layer_a.get_vp_overrides()
        vp_handle = "FEFE"
        vp_overrides.set_color(vp_handle, 6)
        assert vp_overrides.get_color(vp_handle) == 6

    @pytest.mark.parametrize(
        "value", [300, ezdxf.colors.BYLAYER, ezdxf.colors.BYBLOCK]
    )
    def test_invalid_color_raises_value_error(self, layer_a, value):
        vp_overrides = layer_a.get_vp_overrides()
        with pytest.raises(ValueError):
            vp_overrides.set_color("FEFE", value)

    def test_set_rgb_override(self, layer_a):
        vp_handle = "FEFE"
        vp_overrides = layer_a.get_vp_overrides()
        vp_overrides.set_rgb(vp_handle, (1, 2, 3))
        assert vp_overrides.get_rgb(vp_handle) == (1, 2, 3)

    @pytest.fixture
    def layer_rgb(self, doc):
        yield doc.layers.add("LAYER_RGB")
        doc.layers.discard("LAYER_RGB")

    def test_remove_rgb_by_override(self, layer_rgb):
        layer_rgb.rgb = (1, 2, 3)
        vp_handle = "FEFE"
        vp_overrides = layer_rgb.get_vp_overrides()
        vp_overrides.set_rgb(vp_handle, None)
        assert vp_overrides.get_rgb(vp_handle) is None

    def test_set_transparency_override(self, layer_a):
        vp_handle = "FEFE"
        vp_overrides = layer_a.get_vp_overrides()
        vp_overrides.set_transparency(vp_handle, 0.5)
        assert vp_overrides.get_transparency(vp_handle) == 0.5

    def test_set_linetype_override(self, layer_a):
        vp_handle = "FEFE"
        vp_overrides = layer_a.get_vp_overrides()
        vp_overrides.set_linetype(vp_handle, "DASHED")
        assert vp_overrides.get_linetype(vp_handle) == "DASHED"

    def test_linetype_without_table_entry_raises_value_error(self, layer_a):
        vp_overrides = layer_a.get_vp_overrides()
        with pytest.raises(ValueError):
            vp_overrides.set_linetype("FEFE", "DoesNotExist")

    def test_set_lineweight_override(self, layer_a):
        vp_handle = "FEFE"
        vp_overrides = layer_a.get_vp_overrides()
        vp_overrides.set_lineweight(vp_handle, 18)
        assert vp_overrides.get_lineweight(vp_handle) == 18

    @pytest.mark.parametrize(
        "value", [300, const.LINEWEIGHT_BYLAYER, const.LINEWEIGHT_BYBLOCK]
    )
    def test_invalid_lineweight_raises_value_error(self, layer_a, value):
        vp_overrides = layer_a.get_vp_overrides()
        with pytest.raises(ValueError):
            vp_overrides.set_lineweight("FEFE", value)

    def test_discard_specific_overrides(self, layer_a):
        vp_overrides = layer_a.get_vp_overrides()
        vp_handle = "FEFE"
        vp_overrides.set_color(vp_handle, 6)
        assert vp_overrides.has_overrides(vp_handle) is True
        vp_overrides.discard(vp_handle)
        assert vp_overrides.has_overrides(vp_handle) is False

    def test_discard_all_overrides(self, layer_a):
        vp_overrides = layer_a.get_vp_overrides()
        vp_overrides.set_color("FEFE", 6)
        vp_overrides.set_color("ABBA", 5)
        assert vp_overrides.has_overrides() is True
        vp_overrides.discard()
        assert vp_overrides.has_overrides() is False

    def test_discard_ignores_none_existing_vp_handles(self, layer_a):
        vp_overrides = layer_a.get_vp_overrides()
        vp_overrides.set_color("FEFE", 6)
        vp_overrides.discard("xyz")  # should not throw an exception
        assert vp_overrides.has_overrides() is True


class TestCommitChanges:
    @pytest.fixture(scope="class")
    def vp1(self, doc):
        layout: Paperspace = doc.layout("Layout1")  # type: ignore
        return layout.add_viewport(
            center=(2.5, 2.5),
            size=(5, 5),
            view_center_point=(7.5, 7.5),
            view_height=10,
            status=2,
        )

    @pytest.fixture(scope="class")
    def vp2(self, doc):
        layout: Paperspace = doc.layout("Layout1")  # type: ignore
        return layout.add_viewport(
            center=(2.5, 2.5),
            size=(5, 5),
            view_center_point=(7.5, 7.5),
            view_height=10,
            status=3,
        )

    @staticmethod
    def set_all_ovr(ovr: LayerOverrides, vp_handle, color, alpha, ltype, lw):
        ovr.set_color(vp_handle, color)
        ovr.set_transparency(vp_handle, alpha)
        ovr.set_linetype(vp_handle, ltype)
        ovr.set_lineweight(vp_handle, lw)

    def test_commit_creates_proper_xdict_structure(self, doc, vp1):
        vp_handle = vp1.dxf.handle
        layer = doc.layers.add("LS_001")
        ovr = layer.get_vp_overrides()
        self.set_all_ovr(ovr, vp_handle, 3, 0.4, "DASHED", 50)
        ovr.commit()

        assert layer.has_extension_dict is True
        xdict = layer.get_extension_dict()
        assert const.OVR_COLOR_KEY in xdict
        assert const.OVR_ALPHA_KEY in xdict
        assert const.OVR_LTYPE_KEY in xdict
        assert const.OVR_LW_KEY in xdict

    def test_load_overrides_for_one_vp(self, doc, vp1):
        vp_handle = vp1.dxf.handle
        layer = doc.layers.add("LS_002")
        ovr = layer.get_vp_overrides()
        self.set_all_ovr(ovr, vp_handle, 3, 0.4, "DASHED", 50)
        ovr.commit()

        ovr2 = layer.get_vp_overrides()
        assert ovr2.has_overrides() is True
        assert ovr2.has_overrides(vp_handle) is True
        assert ovr2.get_color(vp_handle) == 3
        assert ovr2.get_transparency(vp_handle) == pytest.approx(0.4)
        assert ovr2.get_linetype(vp_handle) == "DASHED"
        assert ovr2.get_lineweight(vp_handle) == 50

    def test_load_overrides_for_two_vp(self, doc, vp1, vp2):
        h1 = vp1.dxf.handle
        h2 = vp2.dxf.handle
        layer = doc.layers.add("LS_003")
        ovr = layer.get_vp_overrides()
        self.set_all_ovr(ovr, h1, 3, 0.4, "DASHED", 50)
        self.set_all_ovr(ovr, h2, 4, 0.2, "CENTER", 35)
        ovr.commit()

        ovr2 = layer.get_vp_overrides()
        assert ovr2.has_overrides() is True
        assert ovr2.has_overrides(h1) is True
        assert ovr2.get_color(h1) == 3
        assert ovr2.get_transparency(h1) == pytest.approx(0.4)
        assert ovr2.get_linetype(h1) == "DASHED"
        assert ovr2.get_lineweight(h1) == 50

        assert ovr2.has_overrides(h2) is True
        assert ovr2.get_color(h2) == 4
        assert ovr2.get_transparency(h2) == pytest.approx(0.2)
        assert ovr2.get_linetype(h2) == "CENTER"
        assert ovr2.get_lineweight(h2) == 35

    def test_commit_no_overrides_creates_no_xdict(self, doc):
        layer = doc.layers.add("LS_004")
        ovr = layer.get_vp_overrides()
        ovr.commit()
        assert layer.has_extension_dict is False

    def test_resetting_overrides_to_default_removes_xdict_entry(self, doc, vp1):
        vp_handle = vp1.dxf.handle
        layer = doc.layers.add("LS_005")
        default_color = layer.get_color()
        ovr = layer.get_vp_overrides()
        ovr.set_color(vp_handle, 3)
        ovr.commit()
        xdict = layer.get_extension_dict()

        # OVR_COLOR_KEY exist:
        assert const.OVR_COLOR_KEY in xdict

        # reset to default color:
        ovr.set_color(vp_handle, default_color)
        ovr.commit()

        # OVR_COLOR_KEY is removed:
        assert layer.has_extension_dict is True, "expected xdict still exist"
        assert const.OVR_COLOR_KEY not in xdict, "layer has no color overrides"


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2018-2024, Manfred Moitzi
# License: MIT License
# legacy.py code is replaced by numpy - exists just for benchmarking, testing and nostalgia
from __future__ import annotations
from typing import Iterable, cast
from itertools import repeat
import reprlib

from .linalg import Matrix, MatrixData, NDArray, Solver

__all__ = [
    "gauss_vector_solver",
    "gauss_matrix_solver",
    "gauss_jordan_solver",
    "gauss_jordan_inverse",
    "LUDecomposition",
]


def copy_float_matrix(A) -> MatrixData:
    if isinstance(A, Matrix):
        A = A.matrix
    return [[float(v) for v in row] for row in A]


def gauss_vector_solver(A: MatrixData | NDArray, B: Iterable[float]) -> list[float]:
    """Solves the linear equation system given by a nxn Matrix A . x = B,
    right-hand side quantities as vector B with n elements by the
    `Gauss-Elimination`_ algorithm, which is faster than the `Gauss-Jordan`_
    algorithm. The speed improvement is more significant for solving multiple
    right-hand side quantities as matrix at once.

    Reference implementation for error checking.

    Args:
        A: matrix [[a11, a12, ..., a1n], [a21, a22, ..., a2n], [a21, a22, ..., a2n],
            ... [an1, an2, ..., ann]]
        B: vector [b1, b2, ..., bn]

    Returns:
        vector as list of floats

    Raises:
        ZeroDivisionError: singular matrix

    """
    # copy input data
    A = copy_float_matrix(A)
    B = list(B)
    num = len(A)
    if len(A[0]) != num:
        raise ValueError("A square nxn matrix A is required.")
    if len(B) != num:
        raise ValueError(
            "Item count of vector B has to be equal to matrix A row count."
        )

    # inplace modification of A & B
    _build_upper_triangle(A, B)
    return _backsubstitution(A, B)


def gauss_matrix_solver(A: MatrixData | NDArray, B: MatrixData | NDArray) -> Matrix:
    """Solves the linear equation system given by a nxn Matrix A . x = B,
    right-hand side quantities as nxm Matrix B by the `Gauss-Elimination`_
    algorithm, which is faster than the `Gauss-Jordan`_ algorithm.

    Reference implementation for error checking.

    Args:
        A: matrix [[a11, a12, ..., a1n], [a21, a22, ..., a2n], [a21, a22, ..., a2n],
            ... [an1, an2, ..., ann]]
        B: matrix [[b11, b12, ..., b1m], [b21, b22, ..., b2m], ... [bn1, bn2, ..., bnm]]

    Returns:
        matrix as :class:`Matrix` object

    Raises:
        ZeroDivisionError: singular matrix

    """
    # copy input data
    matrix_a = copy_float_matrix(A)
    matrix_b = copy_float_matrix(B)
    num = len(matrix_a)
    if len(matrix_a[0]) != num:
        raise ValueError("A square nxn matrix A is required.")
    if len(matrix_b) != num:
        raise ValueError("Row count of matrices A and B has to match.")

    # inplace modification of A & B
    _build_upper_triangle(matrix_a, matrix_b)

    columns = Matrix(matrix=matrix_b).cols()
    result = Matrix()
    for col in columns:
        result.append_col(_backsubstitution(matrix_a, col))

    return result


def _build_upper_triangle(A: MatrixData, B: list) -> None:
    """Build upper triangle for backsubstitution. Modifies A and B inplace!

    Args:
         A: row major matrix
         B: vector of floats or row major matrix

    """
    num = len(A)
    try:
        b_col_count = len(B[0])
    except TypeError:
        b_col_count = 1

    for i in range(0, num):
        # Search for maximum in this column
        max_element = abs(A[i][i])
        max_row = i
        for row in range(i + 1, num):
            value = abs(A[row][i])
            if value > max_element:
                max_element = value
                max_row = row

        # Swap maximum row with current row
        A[max_row], A[i] = A[i], A[max_row]
        B[max_row], B[i] = B[i], B[max_row]

        # Make all rows below this one 0 in current column
        for row in range(i + 1, num):
            c = -A[row][i] / A[i][i]
            for col in range(i, num):
                if i == col:
                    A[row][col] = 0
                else:
                    A[row][col] += c * A[i][col]
            if b_col_count == 1:
                B[row] += c * B[i]
            else:
                for col in range(b_col_count):
                    B[row][col] += c * B[i][col]


def _backsubstitution(A: MatrixData, B: list[float]) -> list[float]:
    """Solve equation A . x = B for an upper triangular matrix A by
    backsubstitution.

    Args:
        A: row major matrix
        B: vector of floats

    """
    num = len(A)
    x = [0.0] * num
    for i in range(num - 1, -1, -1):
        x[i] = B[i] / A[i][i]
        for row in range(i - 1, -1, -1):
            B[row] -= A[row][i] * x[i]
    return x


def gauss_jordan_solver(
    A: MatrixData | NDArray, B: MatrixData | NDArray
) -> tuple[Matrix, Matrix]:
    """Solves the linear equation system given by a nxn Matrix A . x = B,
    right-hand side quantities as nxm Matrix B by the `Gauss-Jordan`_ algorithm,
    which is the slowest of all, but it is very reliable. Returns a copy of the
    modified input matrix `A` and the result matrix `x`.

    Internally used for matrix inverse calculation.

    Args:
        A: matrix [[a11, a12, ..., a1n], [a21, a22, ..., a2n], [a21, a22, ..., a2n],
            ... [an1, an2, ..., ann]]
        B: matrix [[b11, b12, ..., b1m], [b21, b22, ..., b2m], ... [bn1, bn2, ..., bnm]]

    Returns:
        2-tuple of :class:`Matrix` objects

    Raises:
        ZeroDivisionError: singular matrix

    """
    # copy input data
    matrix_a = copy_float_matrix(A)
    matrix_b = copy_float_matrix(B)

    n = len(matrix_a)
    m = len(matrix_b[0])

    if len(matrix_a[0]) != n:
        raise ValueError("A square nxn matrix A is required.")
    if len(matrix_b) != n:
        raise ValueError("Row count of matrices A and B has to match.")

    icol = 0
    irow = 0
    col_indices = [0] * n
    row_indices = [0] * n
    ipiv = [0] * n

    for i in range(n):
        big = 0.0
        for j in range(n):
            if ipiv[j] != 1:
                for k in range(n):
                    if ipiv[k] == 0:
                        if abs(matrix_a[j][k]) >= big:
                            big = abs(matrix_a[j][k])
                            irow = j
                            icol = k

        ipiv[icol] += 1
        if irow != icol:
            matrix_a[irow], matrix_a[icol] = matrix_a[icol], matrix_a[irow]
            matrix_b[irow], matrix_b[icol] = matrix_b[icol], matrix_b[irow]

        row_indices[i] = irow
        col_indices[i] = icol

        pivinv = 1.0 / matrix_a[icol][icol]
        matrix_a[icol][icol] = 1.0
        matrix_a[icol] = [v * pivinv for v in matrix_a[icol]]
        matrix_b[icol] = [v * pivinv for v in matrix_b[icol]]
        for row in range(n):
            if row == icol:
                continue
            dum = matrix_a[row][icol]
            matrix_a[row][icol] = 0.0
            for col in range(n):
                matrix_a[row][col] -= matrix_a[icol][col] * dum
            for col in range(m):
                matrix_b[row][col] -= matrix_b[icol][col] * dum

    for i in range(n - 1, -1, -1):
        irow = row_indices[i]
        icol = col_indices[i]
        if irow != icol:
            for _row in matrix_a:
                _row[irow], _row[icol] = _row[icol], _row[irow]
    return Matrix(matrix=matrix_a), Matrix(matrix=matrix_b)


def gauss_jordan_inverse(A: MatrixData) -> Matrix:
    """Returns the inverse of matrix `A` as :class:`Matrix` object.

    .. hint::

        For small matrices (n<10) is this function faster than
        LUDecomposition(m).inverse() and as fast even if the decomposition is
        already done.

    Raises:
        ZeroDivisionError: singular matrix

    """
    if isinstance(A, Matrix):
        matrix_a = A.matrix
    else:
        matrix_a = list(A)
    nrows = len(matrix_a)
    return gauss_jordan_solver(matrix_a, list(repeat([0.0], nrows)))[0]


class LUDecomposition(Solver):
    """Represents a `LU decomposition`_ matrix of A, raise :class:`ZeroDivisionError`
    for a singular matrix.

    This algorithm is a little bit faster than the `Gauss-Elimination`_
    algorithm using CPython and much faster when using pypy.

    The :attr:`LUDecomposition.matrix` attribute gives access to the matrix data
    as list of rows like in the :class:`Matrix` class, and the :attr:`LUDecomposition.index`
    attribute gives access to the swapped row indices.

    Args:
        A: matrix [[a11, a12, ..., a1n], [a21, a22, ..., a2n], [a21, a22, ..., a2n],
            ... [an1, an2, ..., ann]]

    raises:
        ZeroDivisionError: singular matrix

    """

    __slots__ = ("matrix", "index", "_det")

    def __init__(self, A: MatrixData | NDArray):
        lu: MatrixData = copy_float_matrix(A)
        n: int = len(lu)
        det: float = 1.0
        index: list[int] = []

        # find max value for each row, raises ZeroDivisionError for singular matrix!
        scaling: list[float] = [1.0 / max(abs(v) for v in row) for row in lu]

        for k in range(n):
            big: float = 0.0
            imax: int = k
            for i in range(k, n):
                temp: float = scaling[i] * abs(lu[i][k])
                if temp > big:
                    big = temp
                    imax = i

            if k != imax:
                for col in range(n):
                    temp = lu[imax][col]
                    lu[imax][col] = lu[k][col]
                    lu[k][col] = temp

                det = -det
                scaling[imax] = scaling[k]

            index.append(imax)
            for i in range(k + 1, n):
                temp = lu[i][k] / lu[k][k]
                lu[i][k] = temp
                for j in range(k + 1, n):
                    lu[i][j] -= temp * lu[k][j]

        self.index: list[int] = index
        self.matrix: MatrixData = lu
        self._det: float = det

    def __str__(self) -> str:
        return str(self.matrix)

    def __repr__(self) -> str:
        return f"{self.__class__} {reprlib.repr(self.matrix)}"

    @property
    def nrows(self) -> int:
        """Count of matrix rows (and cols)."""
        return len(self.matrix)

    def solve_vector(self, B: Iterable[float]) -> list[float]:
        """Solves the linear equation system given by the nxn Matrix A . x = B,
        right-hand side quantities as vector B with n elements.

        Args:
            B: vector [b1, b2, ..., bn]

        Returns:
            vector as list of floats

        """
        X: list[float] = [float(v) for v in B]
        lu: MatrixData = self.matrix
        index: list[int] = self.index
        n: int = self.nrows
        ii: int = 0

        if len(X) != n:
            raise ValueError(
                "Item count of vector B has to be equal to matrix row count."
            )

        for i in range(n):
            ip: int = index[i]
            sum_: float = X[ip]
            X[ip] = X[i]
            if ii != 0:
                for j in range(ii - 1, i):
                    sum_ -= lu[i][j] * X[j]
            elif sum_ != 0.0:
                ii = i + 1
            X[i] = sum_

        for row in range(n - 1, -1, -1):
            sum_ = X[row]
            for col in range(row + 1, n):
                sum_ -= lu[row][col] * X[col]
            X[row] = sum_ / lu[row][row]
        return X

    def solve_matrix(self, B: MatrixData | NDArray) -> Matrix:
        """Solves the linear equation system given by the nxn Matrix A . x = B,
        right-hand side quantities as nxm Matrix B.

        Args:
            B: matrix [[b11, b12, ..., b1m], [b21, b22, ..., b2m],
                ... [bn1, bn2, ..., bnm]]

        Returns:
            matrix as :class:`Matrix` object

        """
        if not isinstance(B, Matrix):
            matrix_b = Matrix(matrix=[list(row) for row in B])
        else:
            matrix_b = cast(Matrix, B)

        if matrix_b.nrows != self.nrows:
            raise ValueError("Row count of self and matrix B has to match.")

        return Matrix(
            matrix=[self.solve_vector(col) for col in matrix_b.cols()]
        ).transpose()

    def inverse(self) -> Matrix:
        """Returns the inverse of matrix as :class:`Matrix` object,
        raise :class:`ZeroDivisionError` for a singular matrix.

        """
        return self.solve_matrix(Matrix.identity(shape=(self.nrows, self.nrows)).matrix)

    def determinant(self) -> float:
        """Returns the determinant of matrix, raises :class:`ZeroDivisionError`
        if matrix is singular.

        """
        det: float = self._det
        lu: MatrixData = self.matrix
        for i in range(self.nrows):
            det *= lu[i][i]
        return det



# Copyright (c) 2018-2020 Manfred Moitzi
# License: MIT License
from ezdxf.math.bspline import (
    open_uniform_knot_vector,
    uniform_knot_vector,
    required_knot_values,
)

open_uniform_order2 = [0.0, 0.0, 1.0, 2.0, 3.0, 4.0, 4.0]
open_uniform_order3 = [0.0, 0.0, 0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 5.0, 5.0]
open_uniform_order4 = [
    0.0,
    0.0,
    0.0,
    0.0,
    1.0,
    2.0,
    3.0,
    4.0,
    5.0,
    6.0,
    6.0,
    6.0,
    6.0,
]


def _normalize(values, max_value):
    return [v / max_value for v in values]


def test_open_uniform_knot_order_2():
    result = open_uniform_knot_vector(5, 2, normalize=False)
    assert len(result) == required_knot_values(5, 2)
    assert result == open_uniform_order2
    assert open_uniform_knot_vector(5, 2, normalize=True) == _normalize(
        open_uniform_order2, 4.0
    )


def test_open_uniform_knot_order_3():
    result = open_uniform_knot_vector(7, 3, normalize=False)
    assert len(result) == required_knot_values(7, 3)
    assert result == open_uniform_order3
    assert open_uniform_knot_vector(7, 3, normalize=True) == _normalize(
        open_uniform_order3, 5.0
    )


def test_open_uniform_knot_order_4():
    result = open_uniform_knot_vector(9, 4, normalize=False)
    assert len(result) == required_knot_values(9, 4)
    assert result == open_uniform_order4
    assert open_uniform_knot_vector(9, 4, normalize=True) == _normalize(
        open_uniform_order4, 6.0
    )


uniform_order2 = [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0]
uniform_order3 = [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]
uniform_order4 = [
    0.0,
    1.0,
    2.0,
    3.0,
    4.0,
    5.0,
    6.0,
    7.0,
    8.0,
    9.0,
    10.0,
    11.0,
    12.0,
]


def test_uniform_knot_order_2():
    result = uniform_knot_vector(5, 2, normalize=False)
    assert len(result) == required_knot_values(5, 2)
    assert result == uniform_order2
    assert uniform_knot_vector(5, 2, normalize=True) == _normalize(
        uniform_order2, 6.0
    )


def test_uniform_knot_order_3():
    result = uniform_knot_vector(7, 3, normalize=False)
    assert len(result) == required_knot_values(7, 3)
    assert result == uniform_order3
    assert uniform_knot_vector(7, 3, normalize=True) == _normalize(
        uniform_order3, 9.0
    )


def test_uniform_knot_order_4():
    result = uniform_knot_vector(9, 4, normalize=False)
    assert len(result) == required_knot_values(9, 4)
    assert result == uniform_order4
    assert uniform_knot_vector(9, 4, normalize=True) == _normalize(
        uniform_order4, 12.0
    )



#  Copyright (c) 2022, Manfred Moitzi
#  License: MIT License

import pytest
import ezdxf
from ezdxf.audit import Auditor


@pytest.fixture(scope="module")
def doc():
    return ezdxf.new()


def test_duplicate_handle_loading_error(doc):
    msp = doc.modelspace()
    p1 = msp.add_point((0, 0))
    p2 = msp.add_point((0, 0))
    count = len(msp)
    # Simulate loading error with same handle for two entities!
    p2.dxf.handle = p1.dxf.handle
    auditor = Auditor(doc)
    block_record = doc.block_records.get("*Model_Space")
    block_record.audit(auditor)
    assert len(auditor.fixes) == 1
    assert len(msp) == count-1


if __name__ == '__main__':
    pytest.main([__file__])



# Copyright (c) 2024, Manfred Moitzi
# License: MIT License
import pytest

import ezdxf
from ezdxf import revcloud


def test_create_revcloud():
    points = [(0, 0), (1, 0), (1, 1), (0, 1)]
    lw_points = revcloud.points(points, segment_length=0.1, bulge=0.5)
    assert len(lw_points) == 40


@pytest.mark.parametrize(
    "points", [[], [(0, 0)], [(0, 0), (0, 0)], [(0, 0), (1, 0), (0, 0)]]
)
def test_too_few_points_raises_exception(points):
    with pytest.raises(ValueError):
        revcloud.points(points, segment_length=0.1)


def test_too_small_segment_length_raises_exception():
    with pytest.raises(ValueError):
        revcloud.points([(0, 0), (1, 0), (1, 1), (0, 1)], segment_length=0)


def test_add_entity():
    doc = ezdxf.new()
    msp = doc.modelspace()
    # counter-clockwise oriented revision cloud:
    lwp = revcloud.add_entity(msp, [(0, 0), (1, 0), (1, 1), (0, 1)], 0.1)

    assert doc.appids.has_entry(revcloud.REVCLOUD_PROPS)
    assert revcloud.is_revcloud(lwp) is True

    # clockwise oriented revision cloud:
    lwp = revcloud.add_entity(msp, [(0, 0), (0, 1), (1, 1), (1, 0)], 0.1)
    assert revcloud.is_revcloud(lwp) is True


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License
import pytest

from ezdxf import path
from ezdxf.fonts import fonts

FONT = "standard.lff"


@pytest.fixture(scope="module", autouse=True)
def check_libre_cad_fonts_available():
    if not fonts.font_manager.has_font(FONT):
        pytest.skip(reason=f"required LibreCAD '{FONT}' font is not available")


def test_get_font_family_for_lff_files():
    # LibreCAD does not include a replacement font for TXT
    font_face = fonts.font_manager.get_font_face("standard.lff")
    assert font_face.family == "standard"


TOL = 0.001


class TestLibreCadFont:
    @pytest.fixture(scope="class")
    def lff(self):
        return fonts.make_font(FONT, 2.5)

    def test_is_a_lff_font(self, lff):
        assert lff.name == FONT

    def test_space_width(self, lff):
        assert lff.space_width() == pytest.approx(1.875, abs=TOL)

    def test_text_width(self, lff):
        assert lff.text_width("1234") == pytest.approx(7.578981985029027, abs=TOL)

    def test_text_width_ex(self, lff):
        assert lff.text_width_ex("1234", cap_height=3, width_factor=2) == pytest.approx(
            18.189556764069664, abs=TOL
        )

    def test_text_path(self, lff):
        text_path = lff.text_path("1234")

        # fmt: off
        assert text_path.command_codes() == [
            1, 1, 4, 1, 3, 3, 1, 1, 4, 1, 3, 3, 4, 1, 3, 3, 1, 3, 3, 1, 4, 1, 1, 4, 1
        ]
        # fmt: on
        box = text_path.bbox()
        assert box.size.x > 6
        assert box.size.y == pytest.approx(2.5789819850290274, abs=TOL)

    def test_text_path_ex(self, lff):
        text_path = lff.text_path_ex("1234", cap_height=3, width_factor=2)
        box = path.precise_bbox(text_path)
        assert box.size.x > 16
        assert box.size.y == pytest.approx(3, abs=TOL)


def test_low_level_glyph_cache_measures_spaces():
    cache = fonts.font_manager.get_lff_glyph_cache(FONT)
    assert cache.get_text_length("   ", cache.font_measurements.cap_height) == 20.25


def test_map_shx_to_lff():
    assert fonts.map_shx_to_lff("txt") == "standard.lff"
    assert fonts.map_shx_to_lff("txt.shx") == "standard.lff"
    assert fonts.map_shx_to_lff("iso") == "iso.lff"
    assert fonts.map_shx_to_lff("iso.shx") == "iso.lff"
    assert fonts.map_shx_to_lff("unknown.shx") == "unknown.shx"


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2019-2022 Manfred Moitzi
# License: MIT License
import pytest

from ezdxf.entities.block import Block, EndBlk
from ezdxf.lldxf.const import DXF12, DXF2000
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text

TEST_CLASS = Block
TEST_TYPE = "BLOCK"

ENTITY_R12 = """0
BLOCK
5
0
8
0
2
BLOCKNAME
70
0
10
0.0
20
0.0
30
0.0
3
BLOCKNAME
1

"""

ENTITY_R2000 = """0
BLOCK
5
0
330
0
100
AcDbEntity
8
0
100
AcDbBlockBegin
2
BLOCKNAME
70
0
10
0.0
20
0.0
30
0.0
3
BLOCKNAME
1

"""


@pytest.fixture(params=[ENTITY_R12, ENTITY_R2000])
def entity(request):
    return TEST_CLASS.from_text(request.param)


def test_registered():
    from ezdxf.entities.factory import ENTITY_CLASSES

    assert TEST_TYPE in ENTITY_CLASSES


def test_default_init():
    entity = TEST_CLASS()
    assert entity.dxftype() == TEST_TYPE


def test_default_new():
    entity = TEST_CLASS.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "base_point": (1, 2, 3),
        },
    )
    assert entity.dxf.layer == "0"
    assert entity.dxf.base_point == (1, 2, 3)
    assert entity.dxf.base_point.x == 1, "is not Vec3 compatible"
    assert entity.dxf.base_point.y == 2, "is not Vec3 compatible"
    assert entity.dxf.base_point.z == 3, "is not Vec3 compatible"


def test_load_from_text(entity):
    assert entity.dxf.layer == "0"
    assert entity.dxf.base_point == (0, 0, 0)


@pytest.mark.parametrize(
    "txt,ver", [(ENTITY_R2000, DXF2000), (ENTITY_R12, DXF12)]
)
def test_write_block_dxf(txt, ver):
    expected = basic_tags_from_text(txt)
    block = TEST_CLASS.from_text(txt)
    collector = TagCollector(dxfversion=ver, optional=True)
    block.export_dxf(collector)
    assert collector.tags == expected

    collector2 = TagCollector(dxfversion=ver, optional=False)
    block.export_dxf(collector2)
    assert collector.has_all_tags(collector2)


ENDBLK_R12 = "  0\nENDBLK\n  5\n0\n  8\n0\n"

ENDBLK_R2000 = """0
ENDBLK
5
0
330
0
100
AcDbEntity
8
0
100
AcDbBlockEnd
"""


@pytest.mark.parametrize(
    "txt,ver", [(ENDBLK_R2000, DXF2000), (ENDBLK_R12, DXF12)]
)
def test_write_endblk_dxf(txt, ver):
    expected = basic_tags_from_text(txt)
    endblk = EndBlk.from_text(txt)
    collector = TagCollector(dxfversion=ver, optional=True)
    endblk.export_dxf(collector)
    assert collector.tags == expected

    collector2 = TagCollector(dxfversion=ver, optional=False)
    endblk.export_dxf(collector2)
    assert collector.has_all_tags(collector2)


MALFORMED_BLOCK = """0
BLOCK
5
0
8
LY_EZDXF
330
0
100
AcDbEntity
2
BLOCKNAME
70
0
10
1.0
20
2.0
30
3.0
100
AcDbBlockBegin
3
IGNORED_BLOCKNAME
1

"""


def test_load_malformed_block():
    block = Block.from_text(MALFORMED_BLOCK)
    assert block.dxf.layer == "LY_EZDXF"
    assert block.dxf.base_point.isclose((1, 2, 3))
    assert block.dxf.name == "BLOCKNAME"


MALFORMED_ENDBLK = """0
ENDBLK
5
0
330
0
100
AcDbBlockEnd
8
LY_EZDXF
100
AcDbEntity
100
AcDbBlockEnd
"""


def test_load_malformed_endblk():
    endblk = EndBlk.from_text(MALFORMED_ENDBLK)
    assert endblk.dxf.layer == "LY_EZDXF"



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License

import pytest
from ezdxf.addons.hpgl2.tokenizer import pe_encode, pe_decode, fractional_bits


class TestPolylineEncoding:
    def test_8_bit(self):
        s = pe_encode(10525)
        assert s[0] == 121
        assert s[1] == 71
        assert s[2] == 196

    def test_7_bit(self):
        s = pe_encode(10525, base=32)
        assert s[0] == 89
        assert s[1] == 80
        assert s[2] == 115


class TestPolylineDecoding:
    def test_8_bit(self):
        s = pe_encode(10525) + pe_encode(-10525)
        values, index = list(pe_decode(s))
        assert values[0] == 10525
        assert values[1] == -10525
        assert index == len(s)

    def test_7_bit(self):
        s = pe_encode(10525, base=32) + pe_encode(-10525, base=32)
        values, index = pe_decode(s, base=32)
        assert values[0] == 10525
        assert values[1] == -10525
        assert index == len(s)

    def test_8_bit_float(self):
        n = fractional_bits(3)
        s = pe_encode(10.525, frac_bits=n) + pe_encode(-10.525, frac_bits=n)
        values, index = list(pe_decode(s, frac_bits=n))
        assert round(values[0], 3) == 10.525
        assert round(values[1], 3) == -10.525
        assert index == len(s)


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License
from typing import List
import pytest
import ezdxf
from ezdxf.disassemble import recursive_decompose
from ezdxf.entities import Point, Insert


@pytest.fixture(scope="module")
def doc():
    d = ezdxf.new()
    l0 = d.blocks.new("L0")
    build_level_0(l0)
    l1 = d.blocks.new("L1")
    build_nesting_level_1(l1)
    return d


def build_level_0(l0):
    # Block of 4 lines in 4 different colors
    l0.add_line((1, -1), (1, 1), dxfattribs={"color": 1})
    l0.add_line((1, 1), (-1, 1), dxfattribs={"color": 2})
    l0.add_line((-1, 1), (-1, -1), dxfattribs={"color": 3})
    l0.add_line((-1, -1), (1, -1), dxfattribs={"color": 4})


def scale(sx, sy, sz):
    return {
        "xscale": sx,
        "yscale": sy,
        "zscale": sz,
    }


def build_nesting_level_1(l1, name="L0"):
    l1.add_blockref(name, (0, 0), dxfattribs=scale(1, 1, 1))
    l1.add_blockref(name, (3, 0), dxfattribs=scale(-1, 1, 1))
    l1.add_blockref(name, (6, 0), dxfattribs=scale(1, -1, 1))
    l1.add_blockref(name, (9, 0), dxfattribs=scale(-1, -1, 1))
    l1.add_blockref(name, (0, 3), dxfattribs=scale(1, 1, -1))
    l1.add_blockref(name, (3, 3), dxfattribs=scale(-1, 1, -1))
    l1.add_blockref(name, (6, 3), dxfattribs=scale(1, -1, -1))
    l1.add_blockref(name, (9, 3), dxfattribs=scale(-1, -1, -1))


def count(doc, block_names) -> int:
    count = 1
    for name in block_names:
        block = doc.blocks.get(name)
        count *= len(block)
    return count


def test_decompose_block_level_0(doc):
    l0 = doc.blocks.get("L0")
    result = list(recursive_decompose(l0))
    assert len(result) == count(doc, ["L0"])


REFLECTIONS = [(1, 1, 1), (-1, 1, 1), (1, -1, 1), (1, 1, -1)]
NAMES = [
    "normal",
    "reflect-x",
    "reflect-y",
    "reflect-z",
]


@pytest.mark.parametrize("sx,sy,sz", REFLECTIONS, ids=NAMES)
def test_decompose_block_reference_level_0(doc, sx, sy, sz):
    msp = doc.modelspace()
    msp.delete_all_entities()
    msp.add_blockref("L0", (0, 0), dxfattribs=scale(sx, sy, sz))
    result = list(recursive_decompose(msp))
    assert len(result) == count(doc, ["L0"])


def test_decompose_block_level_1(doc):
    l1 = doc.blocks.get("L1")
    types = [e.dxftype() for e in recursive_decompose(l1)]
    assert len(types) == count(doc, ["L0", "L1"])
    assert set(types) == {"LINE"}, "expected only LINES"


@pytest.mark.parametrize("sx,sy,sz", REFLECTIONS, ids=NAMES)
def test_decompose_block_reference_level_1(doc, sx, sy, sz):
    msp = doc.modelspace()
    msp.delete_all_entities()
    msp.add_blockref("L1", (0, 0), dxfattribs=scale(sx, sy, sz))
    types = [e.dxftype() for e in recursive_decompose(msp)]
    assert len(types) == count(doc, ["L0", "L1"])
    assert set(types) == {"LINE"}, "expected only LINES"


def test_decompose_minsert_level_1(doc):
    nrows = 2
    ncols = 2
    expected_count = count(doc, ["L0", "L1"]) * nrows * ncols

    msp = doc.modelspace()
    msp.delete_all_entities()
    msp.add_blockref(
        "L1",
        (0, 0),
        dxfattribs={
            "row_count": nrows,
            "row_spacing": 5,
            "column_count": ncols,
            "column_spacing": 5,
        },
    )
    types = [e.dxftype() for e in recursive_decompose(msp)]
    assert len(types) == expected_count
    assert set(types) == {"LINE"}, "expected only LINES"


class TestSourceBlockReferences:
    @pytest.fixture(scope="class")
    def doc(self):
        doc_ = ezdxf.new()
        blk0 = doc_.blocks.new("BLK0")
        blk1 = doc_.blocks.new("BLK1")
        blk2 = doc_.blocks.new("BLK2")

        blk2.add_point((2, 1))
        blk2.add_point((2, 2))
        blk1.add_point((1, 1))
        blk1.add_point((1, 2))
        # block reference 2
        blk1.add_blockref("BLK2", (2, 2))
        blk0.add_point((0, 1))
        blk0.add_point((0, 2))
        # block reference 1
        blk0.add_blockref("BLK1", (1, 1))
        # block reference 0
        doc_.modelspace().add_blockref("BLK0", (0, 0))
        return doc_

    @pytest.fixture(scope="class")
    def entities(self, doc):
        return list(recursive_decompose(doc.modelspace()))

    def test_count_of_expected_virtual_entities(self, entities: List[Point]):
        """The virtual INSERT entities are not returned by the
        recursive_decompose() function.
        """
        assert len(entities) == 6, "expected only 6 POINT entities"

    def test_same_source_block_references_blk0(self, entities):
        """Entities from the same INSERT entity have the same virtual source
        block reference.
        """
        insert0 = entities[0].source_block_reference
        insert1 = entities[1].source_block_reference
        assert isinstance(insert0, Insert)
        assert insert0 is insert1
        assert insert0.dxf.name == "BLK0"

    def test_same_source_block_references_blk1(self, entities):
        """Entities from the same INSERT entity have the same virtual source
        block reference.
        """
        insert0 = entities[2].source_block_reference
        insert1 = entities[3].source_block_reference
        assert insert0 is insert1
        assert insert0.dxf.name == "BLK1"

    def test_same_source_block_references_blk2(self, entities):
        """Entities from the same INSERT entity have the same virtual source
        block reference.
        """
        insert0 = entities[4].source_block_reference
        insert1 = entities[5].source_block_reference
        assert insert0 is insert1
        assert insert0.dxf.name == "BLK2"

    def test_link_structure_of_virtual_block_references(self, entities):
        """It is possible to trace back the nesting structure of the block
        references by the source_block_reference property of the virtual
        INSERT entities.
        """
        blkref0 = entities[0].source_block_reference
        blkref1 = entities[2].source_block_reference
        blkref2 = entities[4].source_block_reference
        # blkref2 resides in blkref1
        assert blkref2.source_block_reference is blkref1
        # and blkref1 resides in blkref0
        assert blkref1.source_block_reference is blkref0
        # and blkref0 has no source_block_reference
        assert blkref0.source_block_reference is None


if __name__ == "__main__":
    pytest.main([__file__])
